##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# standard libraries imports
import os, sys

# external libraries imports
import numpy as np

# fullrmc library imports
from fullrmc.Globals import LOGGER, maxint
from fullrmc.Engine import Engine
from fullrmc.Constraints.BondConstraints import BondConstraint
from fullrmc.Constraints.AngleConstraints import BondsAngleConstraint


##########################################################################################
##################################  SHUT DOWN LOGGING  ###################################
LOGGER.set_minimum_level(maxint, stdoutFlag=True, fileFlag=True)


##########################################################################################
#####################################  CREATE ENGINE  ####################################
pdbPath = "system.pdb"
ENGINE = Engine(path=None)
ENGINE.set_pdb(pdbPath)

# add constraints
B_CONSTRAINT  = BondConstraint()
BA_CONSTRAINT = BondsAngleConstraint()
ENGINE.add_constraints([B_CONSTRAINT, BA_CONSTRAINT])
B_CONSTRAINT.create_bonds_by_definition( bondsDefinition={"TIP": [('OH2' ,'H1' , 0.8, 1.1),
                                                                  ('OH2' ,'H2' , 0.8, 1.1)] })
BA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"TIP": [ ('OH2'  ,'H1' ,'H2' , 80, 120)] })

# set TranslationGenerator move generators amplitude
[g.moveGenerator.set_amplitude(0.025) for g in ENGINE.groups]


##########################################################################################
####################################  DIFFERENT RUNS  ####################################
def run_normal(nsteps, xyzPath):
    BA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"TIP": [ ('OH2'  ,'H1' ,'H2' , 80, 120)] })
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps*2, xyzFrequency=1, xyzPath=xyzPath, restartPdb=None)

def run_squeezed(nsteps, xyzPath):
    BA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"TIP": [ ('OH2'  ,'H1' ,'H2' , 30, 40)] })
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps*2, xyzFrequency=1, xyzPath=xyzPath, restartPdb=None)

def run_streched(nsteps, xyzPath):
    BA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"TIP": [ ('OH2'  ,'H1' ,'H2' , 160, 170)] })
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps*2, xyzFrequency=1, xyzPath=xyzPath, restartPdb=None)


##########################################################################################
#####################################  RUN SIMULATION  ###################################
xyzPath ="trajectory.xyz"
if os.path.isfile(xyzPath): os.remove(xyzPath)
run_normal(2000,   xyzPath)
run_squeezed(2000, xyzPath)
run_streched(2000, xyzPath)


##########################################################################################
##################################  VISUALIZE SIMULATION  ################################
ENGINE.set_pdb(pdbPath)
ENGINE.visualize( commands = ["trajectory.xyz"],
                  boxWidth=0, bgColor="white",
                  representationParams='CPK 1.0 0.2 50 50',
                  otherParams=["label add Angles 0/1 0/0 0/2",
                               "label textsize 1.5",
                               "label textthickness 2",
                               "color Labels Angles black",] )



"""
Constraint contains parent classes for all constraints.
A Constraint is used to set certain rules for the stochastic engine to
evolve the atomic system. Therefore it has become possible to fully
customize and set any possibly imaginable rule.

.. inheritance-diagram:: fullrmc.Core.Constraint
    :parts: 1
"""
# standard libraries imports
from __future__ import print_function
import os, inspect, copy, uuid, re, itertools, shutil
from random import random as randfloat

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import ListenerBase, is_number, is_integer, get_path
from ..Core.Collection import _AtomsCollector, reset_if_collected_out_of_date
from ..Core.Collection import get_caller_frames

class Constraint(ListenerBase):
    """ A constraint is used to direct the evolution of the atomic
    configuration towards the desired and most meaningful one.
    """
    def __init__(self):
        # init ListenerBase
        super(Constraint, self).__init__()
        # set engine
        self.__engine = None
        # set used flag
        self.set_used(True)
        # initialize variance squared
        self.__varianceSquared = 1
        # initialize atoms collector with datakeys to None. so it must be set in all subclasses.
        self._atomsCollector = _AtomsCollector(self, dataKeys=None)
        # initialize data
        self.__initialize_constraint()
        # computation cost
        self.set_computation_cost(0)
        # set frame data
        FRAME_DATA      = ('_Constraint__state', '_Constraint__used',
                           '_Constraint__tried', '_Constraint__accepted',
                           '_Constraint__varianceSquared','_Constraint__standardError',
                           '_Constraint__originalData', '_Constraint__data',
                           '_Constraint__activeAtomsDataBeforeMove', '_Constraint__activeAtomsDataAfterMove',
                           '_Constraint__amputationData', '_Constraint__afterMoveStandardError',
                           '_Constraint__amputationStandardError', '_Constraint__computationCost',
                           '_atomsCollector')
        RUNTIME_DATA    = ('_Constraint__state','_Constraint__tried', '_Constraint__accepted',
                           '_Constraint__varianceSquared','_Constraint__standardError','_Constraint__data',
                           '_atomsCollector',)
        object.__setattr__(self, 'FRAME_DATA',      tuple(FRAME_DATA)      )
        object.__setattr__(self, 'RUNTIME_DATA',    tuple(RUNTIME_DATA)    )


    def __setattr__(self, name, value):
        if name in ('FRAME_DATA','RUNTIME_DATA',):
            raise LOGGER.error("Setting '%s' is not allowed."%name)
        else:
            object.__setattr__(self, name, value)

    def __getstate__(self):
        state = {}
        for k in self.__dict__:
            if k in self.FRAME_DATA:
                continue
            else:
                state[k] = self.__dict__[k]
        # remove repository from engine
        #if state.get('_Constraint__engine', None) is not None:
        #    state['_Constraint__engine']._Engine__repository = None
        #print(self.__class__.__name__, '__getstate__')
        return state

    #def __setstate(self, state):
    #    #print(self.__class__.__name__, '__setstate__')
    #    self.__dict__ = state

    def _set_engine(self, engine):
        assert self.__engine is None, LOGGER.error("Re-setting constraint engine is not allowed.")
        from fullrmc.Engine import Engine
        assert isinstance(engine, Engine),LOGGER.error("Engine must be a fullrmc Engine instance")
        self.__engine = engine
        # set constraint unique id
        names = [c.constraintName for c in engine.constraints]
        idx = 0
        while True:
            name = self.__class__.__name__ + "_%i"%idx
            if name not in names:
                self.__constraintName = name
                break
            else:
                idx += 1
        # reset flags
        # resetting is commented because changing engine is not allowed anymore
        # and there is no need to reset_constraint anymore since at this point
        # all flag are not altered.
        #self.reset_constraint(reinitialize=False, flags=True, data=False)

    def __initialize_constraint(self, frame=None):
        usedIncluded, frame, allFrames = get_caller_frames(engine=self.engine,
                                                           frame=frame,
                                                           subframeToAll=False,
                                                           caller="%s.%s"%(self.__class__.__name__,inspect.stack()[0][3]) )
        # initialize flags
        if usedIncluded:
        #if frame is None:
            self.__state     = None
            self.__tried     = 0
            self.__accepted  = 0
            # initialize data
            self.__originalData              = None
            self.__data                      = None
            self.__activeAtomsDataBeforeMove = None
            self.__activeAtomsDataAfterMove  = None
            self.__amputationData            = None
            # initialize standard error
            self.__standardError           = None
            self.__afterMoveStandardError  = None
            self.__amputationStandardError = None
            ## reset atoms collector # ADDED 2017-JAN-08
            self._atomsCollector.reset()
            self._on_collector_reset()
            if self.engine is not None and len(self._atomsCollector.dataKeys):
                for realIndex in self.engine._atomsCollector.indexes:
                    self._on_collector_collect_atom(realIndex=realIndex)
        # loop all frames
        usedFrame = self.usedFrame
        for frm in allFrames:
            ac   = self._atomsCollector
            if frm != usedFrame:
                ac   = self._get_repository().pull(os.path.join(frm, 'constraints', self.__constraintName,'_atomsCollector'))
                this = copy.deepcopy(self)
                this._atomsCollector = ac
                this._atomsCollector.reset()
                this._on_collector_reset()
                if this.engine is not None and len(this._atomsCollector.dataKeys):
                    for realIndex in this.engine._atomsCollector.indexes:
                        this._on_collector_collect_atom(realIndex=realIndex)
            # dump into repostory
            self._dump_to_repository({'_Constraint__state'                    : None,
                                      '_Constraint__tried'                    : 0,
                                      '_Constraint__accepted'                 : 0,
                                      '_Constraint__originalData'             : None,
                                      '_Constraint__data'                     : None,
                                      '_Constraint__activeAtomsDataBeforeMove': None,
                                      '_Constraint__activeAtomsDataAfterMove' : None,
                                      '_Constraint__amputationData'           : None,
                                      '_Constraint__standardError'            : None,
                                      '_Constraint__afterMoveStandardError'   : None,
                                      '_Constraint__amputationStandardError'  : None,
                                      '_atomsCollector'                       : ac}, frame=frm)


    @property
    def constraintId(self):
        """Constraint unique ID create at instanciation time."""
        return self.listenerId

    @property
    def constraintName(self):
        """ Constraints unique name in engine given when added to engine."""
        return self.__constraintName

    @property
    def engine(self):
        """ Stochastic fullrmc's engine instance."""
        return self.__engine

    @property
    def usedFrame(self):
        """Get used frame in engine. If None then engine is not defined yet"""
        usedFrame = None
        if self.__engine is not None:
            usedFrame = self.__engine.usedFrame
        return usedFrame


    @property
    def computationCost(self):
        """ Computation cost number."""
        return self.__computationCost

    @property
    def state(self):
        """ Constraint's state."""
        return self.__state

    @property
    def tried(self):
        """ Constraint's number of tried moves."""
        return self.__tried

    @property
    def accepted(self):
        """ Constraint's number of accepted moves."""
        return self.__accepted

    @property
    def used(self):
        """ Constraint's used flag. Defines whether constraint is used
        in the stochastic engine at runtime or set inactive."""
        return self.__used

    @property
    def varianceSquared(self):
        """ Constraint's varianceSquared used in the stochastic engine
        at runtime to calculate the total constraint's standard error."""
        return self.__varianceSquared

    @property
    def standardError(self):
        """ Constraint's standard error value."""
        return self.__standardError

    @property
    def originalData(self):
        """ Constraint's original data calculated upon initialization."""
        return self.__originalData

    @property
    def data(self):
        """ Constraint's current calculated data."""
        return self.__data

    @property
    def activeAtomsDataBeforeMove(self):
        """ Constraint's current calculated data before last move."""
        return self.__activeAtomsDataBeforeMove

    @property
    def activeAtomsDataAfterMove(self):
        """ Constraint's current calculated data after last move."""
        return self.__activeAtomsDataAfterMove

    @property
    def afterMoveStandardError(self):
        """ Constraint's current calculated StandardError after last move."""
        return self.__afterMoveStandardError

    @property
    def amputationData(self):
        """ Constraint's current calculated data after amputation."""
        return self.__amputationData

    @property
    def amputationStandardError(self):
        """ Constraint's current calculated StandardError after amputation."""
        return self.__amputationStandardError

    @property
    def multiframeWeight(self):
        """Get constraint weight towards total in a multiframe system. """
        return FLOAT_TYPE(1.)

    def _apply_multiframe_prior(self, total):
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def _apply_scale_factor(self, total, scaleFactor):
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def _get_repository(self):
        if self.engine is None:
            return None
        else:
            return self.engine._get_repository()

    def _dump_to_repository(self, dataDict, frame=None):
        rep = self._get_repository()
        if rep is None:
            return
        if frame is None:
            frame = self.engine.usedFrame
        cp = os.path.join(frame, 'constraints', self.__constraintName)
        for name in dataDict:
            relativePath = os.path.join(cp,name)
            isRepoFile,fileOnDisk, infoOnDisk, classOnDisk = rep.is_repository_file(relativePath)
            if isRepoFile and fileOnDisk and infoOnDisk and classOnDisk:
                rep.update(value=dataDict[name], relativePath=relativePath)
            else:
                rep.dump(value=dataDict[name], relativePath=relativePath, replace=True)

    def _set_original_data(self, data, frame=None):
        """ Used only by the stochastic engine to set constraint's data as
        initialized for the first time."""
        self.__originalData = data
        # dump to repository
        self._dump_to_repository({'_Constraint__originalData' :self.__originalData}, frame=frame)

    def _runtime_initialize(self):
        """
        This is called once everytime engine.run method is executed.
        It is meant to be used as a final setup call for all constraints.
        """
        pass

    def _runtime_on_step(self):
        """
        This is called at everytime engine.run method main loop step.
        """
        pass

    def is_in_engine(self, engine):
        """
        Get whether constraint is already in defined and added to engine.
        It can be the same exact instance or a repository pulled instance
        of the same constraintId

        :Parameters:
            #. engine (stochastic fullrmc engine): Engine instance.

        :Returns:
            #. result (boolean): Whether constraint exists in engine.
        """
        if self in engine.constraints:
            return True
        elif self.constraintId in [c.constraintId for c in engine.constraints]:
            return True
        return False

    def set_variance_squared(self, value, frame=None):
        """
        Set constraint's variance squared that is used in the computation
        of the total stochastic engine standard error.

        :Parameters:
            #. value (number): Any positive non zero number.
            #. frame (None, string): Target frame name. If None, engine used
               frame is used. If multiframe is given, all subframes will be
               targeted. If subframe is given, all other multiframe subframes
               will be targeted.
        """
        assert is_number(value), LOGGER.error("Variance squared accepted value must be convertible to a number")
        value = FLOAT_TYPE(value)
        assert value>0 , LOGGER.error("Variance squared must be positive non zero number.")
        usedIncluded, frame, allFrames = get_caller_frames(engine=self.engine,
                                                           frame=frame,
                                                           subframeToAll=True,
                                                           caller="%s.%s"%(self.__class__.__name__,inspect.stack()[0][3]) )
        if usedIncluded:
            self.__varianceSquared = value
        for frm in allFrames:
            self._dump_to_repository({'_Constraint__varianceSquared' :value}, frame=frm)


    def set_computation_cost(self, value, frame=None):
        """
        Set constraint's computation cost value. This is used at stochastic
        engine runtime to minimize computations and enhance performance by
        computing less costly constraints first. At every step, constraints
        will be computed in order starting from the less to the most
        computationally costly. Therefore upon rejection of a step because
        of an unsatisfactory rigid constraint, the left un-computed
        constraints at this step are guaranteed to be the most time coslty
        ones.

        :Parameters:
            #. value (number): computation cost.
            #. frame (None, string): Target frame name. If None, engine used
               frame is used. If multiframe is given, all subframes will be
               targeted. If subframe is given, all other multiframe subframes
               will be targeted.
        """
        assert is_number(value), LOGGER.error("computation cost value must be convertible to a number")
        value = FLOAT_TYPE(value)
        usedIncluded, frame, allFrames = get_caller_frames(engine=self.engine,
                                                           frame=frame,
                                                           subframeToAll=True,
                                                           caller="%s.%s"%(self.__class__.__name__,inspect.stack()[0][3]) )
        if usedIncluded:
            self.__computationCost = value
        # dump to repository
        for frm in allFrames:
            self._dump_to_repository({'_Constraint__computationCost' :self.__computationCost}, frame=frm)


    @reset_if_collected_out_of_date # ADDED 2017-JAN-12
    def set_used(self, value, frame=None):
        """
        Set used flag.

        :Parameters:
            #. value (boolean): True to use this constraint in stochastic
               engine runtime.
            #. frame (None, string): Target frame name. If None, engine used
               frame is used. If multiframe is given, all subframes will be
               targeted. If subframe is given, all other multiframe subframes
               will be targeted.
        """
        assert isinstance(value, bool), LOGGER.error("value must be boolean")
        # get used frame
        usedIncluded, frame, allFrames = get_caller_frames(engine=self.engine,
                                                           frame=frame,
                                                           subframeToAll=True,
                                                           caller="%s.%s"%(self.__class__.__name__,inspect.stack()[0][3]) )
        if usedIncluded:
            self.__used = value
        for frm in allFrames:
            self._dump_to_repository({'_Constraint__used' :value}, frame=frm)

    def set_state(self, value):
        """
        Set constraint's state. When constraint's state and stochastic
        engine's state don't match, constraint's data must be re-calculated.

        :Parameters:
            #. value (object): Constraint state value.
        """
        self.__state = value

    def set_tried(self, value):
        """
        Set constraint's number of tried moves.

        :Parameters:
            #. value (integer): Constraint tried moves value.
        """
        try:
            value = float(value)
        except:
            raise Exception(LOGGER.error("tried value must be convertible to a number"))
        assert is_integer(value), LOGGER.error("tried value must be integer")
        assert value>=0, LOGGER.error("tried value must be positive")
        self.__tried = int(value)

    def increment_tried(self):
        """ Increment number of tried moves. """
        self.__tried += 1

    def set_accepted(self, value):
        """
        Set constraint's number of accepted moves.

        :Parameters:
            #. value (integer): Constraint's number of accepted moves.
        """
        try:
            value = float(value)
        except:
            raise Exception(LOGGER.error("accepted value must be convertible to a number"))
        assert is_integer(value), LOGGER.error("accepted value must be integer")
        assert value>=0, LOGGER.error("accepted value must be positive")
        assert value<=self.__tried, LOGGER.error("accepted value can't be bigger than number of tried moves")
        self.__accepted = int(value)

    def increment_accepted(self):
        """ Increment constraint's number of accepted moves. """
        self.__accepted += 1

    def set_standard_error(self, value):
        """
        Set constraint's standardError value.

        :Parameters:
            #. value (number): standard error value.
        """
        self.__standardError = value

    def set_data(self, value):
        """
        Set constraint's data value.

        :Parameters:
            #. value (number): constraint's data.
        """
        self.__data = value

    def set_active_atoms_data_before_move(self, value):
        """
        Set constraint's before move happens active atoms data value.

        :Parameters:
            #. value (number): Data value.
        """
        self.__activeAtomsDataBeforeMove = value

    def set_active_atoms_data_after_move(self, value):
        """
        Set constraint's after move happens active atoms data value.

        :Parameters:
            #. value (number): data value.
        """
        self.__activeAtomsDataAfterMove = value

    def set_after_move_standard_error(self, value):
        """
        Set constraint's standard error value after move happens.

        :Parameters:
            #. value (number): standard error value.
        """
        self.__afterMoveStandardError = value

    def set_amputation_data(self, value):
        """
        Set constraint's after amputation data.

        :Parameters:
            #. value (number): data value.
        """
        self.__amputationData = value

    def set_amputation_standard_error(self, value):
        """
        Set constraint's standardError after amputation.

        :Parameters:
            #. value (number): standard error value.
        """
        self.__amputationStandardError = value

    def reset_constraint(self, reinitialize=True, flags=False, data=False, frame=None):
        """
        Reset constraint.

        :Parameters:
            #. reinitialize (boolean): If set to True, it will override
               the rest of the flags and will completely reinitialize the
               constraint.
            #. flags (boolean): Reset the state, tried and accepted flags
               of the constraint.
            #. data (boolean): Reset the constraints computed data.
            #. frame (None, string): Target frame name. If None, engine used
               frame is used. If multiframe is given, all subframes will be
               targeted. If subframe is given, rest of multiframe subframes
               will not be targeted.
        """
        usedIncluded, frame, allFrames = get_caller_frames(engine=self.engine,
                                                           frame=frame,
                                                           subframeToAll=False,
                                                           caller="%s.%s"%(self.__class__.__name__,inspect.stack()[0][3]) )
        # reinitialize constraint
        if reinitialize:
            flags = False
            data  = False
            self.__initialize_constraint(frame=frame)
        # initialize flags
        if flags:
            #if frame is None:
            if usedIncluded:
                self.__state                  = None
                self.__tried                  = 0
                self.__accepted               = 0
                self.__standardError          = None
                self.__afterMoveStandardError = None
            # dunp to repository
            for frm in allFrames:
                self._dump_to_repository({'_Constraint__state'                  : None,
                                          '_Constraint__tried'                  : 0,
                                          '_Constraint__accepted'               : 0,
                                          '_Constraint__standardError'          : None,
                                          '_Constraint__afterMoveStandardError' : None}, frame=frm)
        # initialize data
        if data:
            #if frame is None:
            if usedIncluded:
                self.__originalData              = None
                self.__data                      = None
                self.__activeAtomsDataBeforeMove = None
                self.__activeAtomsDataAfterMove  = None
                self.__standardError             = None
                self.__afterMoveStandardError    = None
            # dunp to repository
            for frm in allFrames:
                self._dump_to_repository({'_Constraint__originalData'             : None,
                                          '_Constraint__data'                     : None,
                                          '_Constraint__activeAtomsDataBeforeMove': None,
                                          '_Constraint__activeAtomsDataAfterMove' : None,
                                          '_Constraint__standardError'            : None,
                                          '_Constraint__afterMoveStandardError'   : None}, frame=frm)


    def update_standard_error(self):
        """ Compute and set constraint's standard error by calling
        compute_standard_error method and passing constraint's data."""
        self.set_standard_error(self.compute_standard_error(data = self.data))

    def get_constraint_value(self):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def get_constraint_original_value(self):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def compute_standard_error(self):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def compute_data(self):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def compute_before_move(self, realIndexes, relativeIndexes):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def compute_after_move(self, realIndexes, relativeIndexes, movedBoxCoordinates):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def accept_move(self, realIndexes, relativeIndexes):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def reject_move(self, realIndexes, relativeIndexes):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def compute_as_if_amputated(self, realIndex, relativeIndex):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def compute_as_if_inserted(self, realIndex, relativeIndex):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def accept_amputation(self, realIndex, relativeIndex):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def reject_amputation(self, realIndex, relativeIndex):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def accept_insertion(self, realIndex, relativeIndex):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def reject_insertion(self, realIndex, relativeIndex):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def _on_collector_collect_atom(self, realIndex):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def _on_collector_release_atom(self, realIndex):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def _on_collector_reset(self):
        """Method must be overloaded in children classes."""
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def export(self, *args, **kwargs):
        """Method must be overloaded in children classes."""
        LOGGER.warn("%s export method is not implemented"%(self.__class__.__name__))

    def plot(self, *args, **kwargs):
        """Method must be overloaded in children classes."""
        LOGGER.warn("%s plot method is not implemented"%(self.__class__.__name__))


class ExperimentalConstraint(Constraint):
    """
    Experimental constraint is any constraint related to experimental data.

    :Parameters:
        #. engine (None, fullrmc.Engine): Constraint's stochastic engine.
        #. experimentalData (numpy.ndarray, string): Experimental data goiven
           as numpy.ndarray or string path to load data using numpy.loadtxt
           method.
        #. dataWeights (None, numpy.ndarray): Weights array of the same number
           of points of experimentalData used in the constraint's standard
           error computation. Therefore particular fitting emphasis can be
           put on different data points that might be considered as more or less
           important in order to get a reasonable and plausible modal.\n
           If None is given, all data points are considered of the same
           importance in the computation of the constraint's standard error.\n
           If numpy.ndarray is given, all weights must be positive and all
           zeros weighted data points won't contribute to the total
           constraint's standard error. At least a single weight point is
           required to be non-zeros and the weights array will be automatically
           scaled upon setting such as the the sum of all the weights
           is equal to the number of data points.
        #. scaleFactor (number): A normalization scale factor used to normalize
           the computed data to the experimental ones.
        #. adjustScaleFactor (list, tuple): Used to adjust fit or guess
           the best scale factor during stochastic engine runtime.
           It must be a list of exactly three entries.\n
           #. The frequency in number of generated moves of finding the best
              scale factor. If 0 frequency is given, it means that the scale
              factor is fixed.
           #. The minimum allowed scale factor value.
           #. The maximum allowed scale factor value.

    **NB**: If adjustScaleFactor first item (frequency) is 0, the scale factor
    will remain untouched and the limits minimum and maximum won't be checked.

    """
    def __init__(self, experimentalData, dataWeights=None, scaleFactor=1.0, adjustScaleFactor=(0, 0.8, 1.2) ):
        # initialize constraint
        super(ExperimentalConstraint, self).__init__()
        # set plotting default parameters.
        self._plotDefaultParameters = {}
        self._plotDefaultParameters['expParams']    = {'label':"experimental","color":'red','marker':'o','markersize':7.5, 'markevery':1, 'zorder':0}
        self._plotDefaultParameters['totParams']    = {'label':"total", 'color':'black','linewidth':3.0, 'zorder':1}
        self._plotDefaultParameters['noWParams']    = {'label':"total - no window", 'color':'black','linewidth':1.0, 'zorder':1}
        self._plotDefaultParameters['shaParams']    = {'label':"shape function", 'color':'black','linewidth':1.0, 'linestyle':'dashed', 'zorder':2}
        self._plotDefaultParameters['parParams']    = {'linewidth':1.0, 'markevery':5, 'markersize':5, 'zorder':3}
        self._plotDefaultParameters['legendParams'] = {'frameon':False, 'ncol':2, 'loc':'best'}
        self._plotDefaultParameters['xlabelParams'] = None
        self._plotDefaultParameters['ylabelParams'] = None
        # set the constraint's experimental data
        self.__dataWeights      = None
        self._usedDataWeights   = None # this is the same as dataWeights but restricted to given limits
        self.__experimentalData = None
        # init limits
        self.__limits           = None
        self.__limitsIndexStart = None
        self.__limitsIndexEnd   = None
        # set constraint data prior. Prior is got from any other structures
        # contributing to the total constraint. This is used in multiframe
        # modeling e.g. distribution of nanoparticules size
        # prior weight can be different per type of constraint as systems
        # respond differently to different probing techniques.
        self._set_multiframe_prior_and_weight(multiframePrior=None, multiframeWeight=None)
        self.set_experimental_data(experimentalData)
        # set scale factor
        self.set_scale_factor(scaleFactor)
        # set adjust scale factor
        self.set_adjust_scale_factor(adjustScaleFactor)
        # set data weights
        self.set_data_weights(dataWeights)
        # set frame data
        FRAME_DATA = [d for d in self.FRAME_DATA]
        FRAME_DATA.extend(['_ExperimentalConstraint__dataWeights',
                           '_ExperimentalConstraint__experimentalData',
                           '_ExperimentalConstraint__scaleFactor',
                           '_ExperimentalConstraint__multiframePrior',
                           '_ExperimentalConstraint__multiframeWeight',
                           '_ExperimentalConstraint__limits',
                           '_ExperimentalConstraint__limitsIndexStart',
                           '_ExperimentalConstraint__limitsIndexEnd',
                           '_usedDataWeights',
                           '_fittedScaleFactor'] )
        RUNTIME_DATA = [d for d in self.RUNTIME_DATA]
        RUNTIME_DATA.extend( ['_ExperimentalConstraint__scaleFactor',
                              '_ExperimentalConstraint__multiframePrior',
                              '_ExperimentalConstraint__multiframeWeight',
                              '_fittedScaleFactor',] )
        object.__setattr__(self, 'FRAME_DATA',  tuple(FRAME_DATA)   )
        object.__setattr__(self, 'RUNTIME_DATA',tuple(RUNTIME_DATA) )


    def _set_multiframe_prior_and_weight(self, multiframePrior, multiframeWeight):
        """To be used internally. Normally at engine run level after returning all
        result from all frames.
        multiframePrior and multiframeWeight can be both None. If not None,
        multiframePrior must have the same length as model constraint total
        and multiframeWeight must a positive value >=0 and <=1
        """
        if multiframePrior is None:
            assert multiframeWeight is None, LOGGER.error('multiframeWeight must be None if multiframePrior is None')
        else:
            # check multiframePrior
            assert isinstance(multiframePrior, np.ndarray), LOGGER.error('multiframePrior must be None or a numpy ndarray')
            assert multiframePrior.dtype == FLOAT_TYPE, LOGGER.error('multiframePrior numpy ndarray data type must be %s'%FLOAT_TYPE)
            assert multiframePrior.shape == self.__experimentalData.shape, LOGGER.error("multiframePrior must have the same shape as experimental data")
            # check multiframeWeight
            assert multiframeWeight is not None, LOGGER.error('multiframeWeight must not be None if multiframePrior is given')
            try:
                multiframeWeight = FLOAT_TYPE(multiframeWeight)
            except Exception as err:
                raise Exception(LOGGER.error("multiframeWeight must be None or a float"))
            assert multiframeWeight>=0, LOGGER.error("multiframeWeight number must be >=0")
            assert multiframeWeight<=1, LOGGER.error("multiframeWeight number must be <=1")
        self.__multiframePrior  = multiframePrior
        self.__multiframeWeight = multiframeWeight
        self._dump_to_repository({'_ExperimentalConstraint__multiframePrior' : self.__multiframePrior,
                                  '_ExperimentalConstraint__multiframeWeight': self.__multiframeWeight})


    def _set_fitted_scale_factor_value(self, scaleFactor):
        """
        This method is a scaleFactor value without any validity checking.
        Meant to be used internally only.
        """
        if self.__scaleFactor != scaleFactor:
            LOGGER.info("@%s Experimental constraint '%s' scale factor updated from %.6f to %.6f" %(self.engine.usedFrame, self.__class__.__name__, self.__scaleFactor, scaleFactor))
            self.__scaleFactor = scaleFactor

    def __set_limits(self, limits):
        """ for internal use only by ExperimentalConstraint children.
        call as such self._ExperimentalConstraint__set_limits(limits)
        """
        if limits is None:
            self.__limits = (None, None)
        else:
            assert isinstance(limits, (list, tuple)), LOGGER.error("limits must be None or a list")
            limits = list(limits)
            assert len(limits) == 2, LOGGER.error("limits list must have exactly two elements")
            if limits[0] is not None:
                assert is_number(limits[0]), LOGGER.error("if not None, the first limits element must be a number")
                limits[0] = FLOAT_TYPE(limits[0])
                assert limits[0]>=0, LOGGER.error("if not None, the first limits element must be bigger or equal to 0")
            if limits[1] is not None:
                assert is_number(limits[1]), LOGGER.error("if not None, the second limits element must be a number")
                limits[1] = FLOAT_TYPE(limits[1])
                assert limits[1]>0, LOGGER.error("if not None, the second limits element must be a positive number")
            if  limits[0] is not None and limits[1] is not None:
                assert limits[0]<limits[1], LOGGER.error("if not None, the first limits element must be smaller than the second limits element")
            self.__limits = (limits[0], limits[1])
        # get minimumDistance and maximumDistance indexes
        if self.limits[0] is None:
            self.__limitsIndexStart = 0
        else:
            self.__limitsIndexStart = (np.abs(self.__experimentalData[:,0]-self.__limits[0])).argmin()
        if self.limits[1] is None:
            self.__limitsIndexEnd = self.__experimentalData.shape[0]-1
        else:
            self.__limitsIndexEnd = (np.abs(self.__experimentalData[:,0]-self.__limits[1])).argmin()
        # dump to repository
        self._dump_to_repository({'_ExperimentalConstraint__limits'           : self.__limits,
                                  '_ExperimentalConstraint__limitsIndexStart' : self.__limitsIndexStart,
                                  '_ExperimentalConstraint__limitsIndexEnd'   : self.__limitsIndexEnd
                                 })

    @property
    def experimentalData(self):
        """ Experimental data of the constraint. """
        return self.__experimentalData

    @property
    def dataWeights(self):
        """ Experimental data points weight"""
        return self.__dataWeights

    @property
    def multiframeWeight(self):
        """Get constraint weight towards total in a multiframe system. """
        return self.__multiframeWeight

    @property
    def multiframePrior(self):
        """Get constraint multiframe prior array. """
        return self.__multiframePrior

    @property
    def scaleFactor(self):
        """ Constraint's scaleFactor. """
        return self.__scaleFactor

    @property
    def adjustScaleFactor(self):
        """Adjust scale factor tuple."""
        return (self.__adjustScaleFactorFrequency, self.__adjustScaleFactorMinimum, self.__adjustScaleFactorMaximum)

    @property
    def adjustScaleFactorFrequency(self):
        """ Scale factor adjustment frequency. """
        return self.__adjustScaleFactorFrequency

    @property
    def adjustScaleFactorMinimum(self):
        """ Scale factor adjustment minimum number allowed. """
        return self.__adjustScaleFactorMinimum

    @property
    def adjustScaleFactorMaximum(self):
        """ Scale factor adjustment maximum number allowed. """
        return self.__adjustScaleFactorMaximum

    @property
    def limits(self):
        """ Used data X limits."""
        return self.__limits

    @property
    def limitsIndexStart(self):
        """ Used data start index as calculated from limits."""
        return self.__limitsIndexStart

    @property
    def limitsIndexEnd(self):
        """ Used data end index as calculated from limits."""
        return self.__limitsIndexEnd


    def _apply_multiframe_prior(self, total):
        """Given a constraint data total, apply multiframe total prior.

        :Parameters:
            #. total (numpy.ndarray): Constraint total.

        :Returns:
            #. transformed (numpy.ndarray): Prior applied to constraint total.
        """
        # THIS MUST BE CORRECTED TO ACCOMODATE DIFFERENT DATA LIMITS
        # REVISITING set_limits IMPLEMENTATION IS NEEDED TO AUTOMATICALLY
        # ADJUST _apply_multiframe_prior TO DATA LIMITS
        # limitsIndexStart AND limitsIndexEnd MUST BE MOVED TO
        # ExperimentalConstraint
        if self.__multiframeWeight is None:
            return total
        else:
            return self.__multiframePrior + self.__multiframeWeight*total

    def set_scale_factor(self, scaleFactor):
        """
        Set the scale factor. This method doesn't allow specifying frames. It
        will target used frame only.

        :Parameters:
             #. scaleFactor (number): A normalization scale factor used to
                normalize the computed data to the experimental ones.
        """
        assert is_number(scaleFactor), LOGGER.error("scaleFactor must be a number")
        self.__scaleFactor      = FLOAT_TYPE(scaleFactor)
        self._fittedScaleFactor = self.__scaleFactor
        # dump to repository
        self._dump_to_repository({'_ExperimentalConstraint__scaleFactor': self.__scaleFactor})
        ## reset constraint
        self.reset_constraint()

    def set_adjust_scale_factor(self, adjustScaleFactor):
        """
        Set adjust scale factor. This method doesn't allow specifying frames. It
        will target used frame only.

        :Parameters:
            #. adjustScaleFactor (list, tuple): Used to adjust fit or guess
               the best scale factor during stochastic engine runtime.
               It must be a list of exactly three entries.\n
               #. The frequency in number of generated moves of finding the best
                  scale factor. If 0 frequency is given, it means that the scale
                  factor is fixed.
               #. The minimum allowed scale factor value.
               #. The maximum allowed scale factor value.
        """
        assert isinstance(adjustScaleFactor, (list, tuple)), LOGGER.error('adjustScaleFactor must be a list.')
        assert len(adjustScaleFactor) == 3, LOGGER.error('adjustScaleFactor must be a list of exactly three items.')
        freq  = adjustScaleFactor[0]
        minSF = adjustScaleFactor[1]
        maxSF = adjustScaleFactor[2]
        assert is_integer(freq), LOGGER.error("adjustScaleFactor first item (frequency) must be an integer.")
        freq = INT_TYPE(freq)
        assert freq>=0, LOGGER.error("adjustScaleFactor first (frequency) item must be bigger or equal to 0.")
        assert is_number(minSF), LOGGER.error("adjustScaleFactor second item (minimum) must be a number.")
        minSF = FLOAT_TYPE(minSF)
        assert is_number(maxSF), LOGGER.error("adjustScaleFactor third item (maximum) must be a number.")
        maxSF = FLOAT_TYPE(maxSF)
        assert minSF<=maxSF, LOGGER.error("adjustScaleFactor second item (minimum) must be smaller or equal to third second item (maximum).")
        # set values
        self.__adjustScaleFactorFrequency = freq
        self.__adjustScaleFactorMinimum   = minSF
        self.__adjustScaleFactorMaximum   = maxSF
        # dump to repository
        self._dump_to_repository({'_ExperimentalConstraint__adjustScaleFactorFrequency': self.__adjustScaleFactorFrequency,
                                  '_ExperimentalConstraint__adjustScaleFactorMinimum'  : self.__adjustScaleFactorMinimum,
                                  '_ExperimentalConstraint__adjustScaleFactorMaximum'  : self.__adjustScaleFactorMaximum})
        # reset constraint
        self.reset_constraint()

    def _set_adjust_scale_factor_frequency(self, freq):
        """This must never be used externally. It's added to serve RemoveGenerators
         and only used internally upon calling compute_as_if_amputated """
        self.__adjustScaleFactorFrequency = freq

    def set_experimental_data(self, experimentalData):
        """
        Set the constraint's experimental data. This method will raise an error
        if called after adding constraint to stochastic engine.

        :Parameters:
            #. experimentalData (numpy.ndarray, string): Experimental data as
               numpy.ndarray or string path to load data using numpy.loadtxt
               method.
        """
        assert self.engine is None, LOGGER.error("Experimental data must be set before engine is set") # ADDED 2018-11-21
        if isinstance(experimentalData, basestring):
            try:
                experimentalData = np.loadtxt(str(experimentalData), dtype=FLOAT_TYPE)
            except Exception as e:
                raise Exception(LOGGER.error("unable to load experimentalData path '%s' (%s)"%(experimentalData, e)))
        assert isinstance(experimentalData, np.ndarray), LOGGER.error("experimentalData must be a numpy.ndarray or string path to load data using numpy.loadtxt.")
        # check data format
        valid, message = self.check_experimental_data(experimentalData)
        # set experimental data
        if valid:
            self.__experimentalData = experimentalData
        else:
            raise Exception( LOGGER.error("%s"%message) )
        # dump to repository
        self._dump_to_repository({'_ExperimentalConstraint__experimentalData': self.__experimentalData})

    def set_data_weights(self, dataWeights, frame=None):
        """
        Set experimental data points weight. Data weights will be automatically
        normalized.

        :Parameters:
            #. dataWeights (None, numpy.ndarray): Weights array of the same
               number of points of experimentalData used in the constraint's
               standard error computation. Therefore particular fitting
               emphasis can be put on different data points that might be
               considered as more or less important in order to get a
               reasonable and plausible model.\n
               If None is given, all data points are considered of the same
               importance in the computation of the constraint's standard error.\n
               If numpy.ndarray is given, all weights must be positive and all
               zeros weighted data points won't contribute to the total
               constraint's standard error. At least a single weight point is
               required to be non-zeros and the weights array will be
               automatically scaled upon setting such as the the sum of all
               the weights is equal to the number of data points.
            #. frame (None, string): Target frame name. If None, engine used
               frame is used. If multiframe is given, all subframes will be
               targeted. If subframe is given, rest of multiframe subframes
               will not be targeted.
        """
        if dataWeights is not None:
            assert isinstance(dataWeights, (list, tuple, np.ndarray)), LOGGER.error("dataWeights must be None or a numpy array of weights")
            try:
                dataWeights = np.array(dataWeights, dtype=FLOAT_TYPE)
            except Exception as e:
                raise Exception(LOGGER.error("unable to cast dataWeights as a numpy array (%s)"%(e)))
            assert len(dataWeights.shape) == 1, LOGGER.error("dataWeights must be a vector")
            assert len(dataWeights) == self.__experimentalData.shape[0], LOGGER.error("dataWeights must be a of the same length as experimental data")
            assert np.min(dataWeights) >=0, LOGGER.error("dataWeights negative values are not allowed")
            assert np.sum(dataWeights), LOGGER.error("dataWeights must be a non-zero array")
            dataWeights /= FLOAT_TYPE( np.sum(dataWeights) )
            dataWeights *= FLOAT_TYPE( len(dataWeights) )
        # get all frames
        usedIncluded, frame, allFrames = get_caller_frames(engine=self.engine,
                                                           frame=frame,
                                                           subframeToAll=False,
                                                           caller="%s.%s"%(self.__class__.__name__,inspect.stack()[0][3]) )
        if usedIncluded:
            self.__dataWeights = dataWeights
        # dump to repository
        for frm in allFrames:
            self._dump_to_repository({'_ExperimentalConstraint__dataWeights': self.__dataWeights},
                                     frame=frm)
        # set used data weights
        self._set_used_data_weights()

    def _set_used_data_weights(self, limitsIndexStart=None, limitsIndexEnd=None):
        # set used dataWeights
        if self.__dataWeights is None:
            self._usedDataWeights = None
        else:
            if limitsIndexStart is None:
                limitsIndexStart = 0
            if limitsIndexEnd is None:
                limitsIndexEnd = self.__experimentalData.shape[0]
            self._usedDataWeights  = np.copy(self.dataWeights[limitsIndexStart:limitsIndexEnd+1])
            assert np.sum(self._usedDataWeights), LOGGER.error("used points dataWeights are all zero.")
            self._usedDataWeights /= FLOAT_TYPE( np.sum(self._usedDataWeights) )
            self._usedDataWeights *= FLOAT_TYPE( len(self._usedDataWeights) )
        # dump to repository
        if self.engine is not None:
            isNormalFrame, isMultiframe, isSubframe = self.engine.get_frame_category(frame=self.engine.usedFrame)
            if isSubframe:
                LOGGER.usage("Setting experimental data weight for multiframe '%s' subframe. This is not going to automatically propagate to all other subframes."%(self.engine.usedFrame,))
        self._dump_to_repository({'_usedDataWeights': self._usedDataWeights})

    def check_experimental_data(self, experimentalData):
        """
        Checks the constraint's experimental data
        This method must be overloaded in all experimental constraint
        sub-classes.

        :Parameters:
            #. experimentalData (numpy.ndarray): Experimental data numpy.ndarray.
        """
        raise Exception(LOGGER.error("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def fit_scale_factor(self, experimentalData, modelData, dataWeights):
        """
        The best scale factor value is computed by minimizing :math:`E=sM`.\n

        Where:
            #. :math:`E` is the experimental data.
            #. :math:`s` is the scale factor.
            #. :math:`M` is the model constraint data.

        This method doesn't allow specifying frames. It will target used frame
        only.

        :Parameters:
            #. experimentalData (numpy.ndarray): Experimental data.
            #. modelData (numpy.ndarray): Constraint modal data.
            #. dataWeights (None, numpy.ndarray): Data points weights to
               compute the scale factor. If None is given, all data points
               will be considered as having the same weight.

        :Returns:
            #. scaleFactor (number): The new scale factor fit value.

        **NB**: This method won't update the internal scale factor value
        of the constraint. It always computes the best scale factor given
        experimental and atomic model data.
        """
        if dataWeights is None:
            SF = FLOAT_TYPE( np.sum(modelData*experimentalData)/np.sum(modelData**2) )
        else:
            SF = FLOAT_TYPE( np.sum(dataWeights*modelData*experimentalData)/np.sum(modelData**2) )
        SF = max(SF, self.__adjustScaleFactorMinimum)
        SF = min(SF, self.__adjustScaleFactorMaximum)
        return SF

    def get_adjusted_scale_factor(self, experimentalData, modelData, dataWeights):
        """
        Checks if scale factor should be updated according to the given scale
        factor frequency and engine's accepted steps. If adjustment is due,
        a new scale factor will be computed using fit_scale_factor method,
        otherwise the the constraint's scale factor will be returned.

        :Parameters:
            #. experimentalData (numpy.ndarray): the experimental data.
            #. modelData (numpy.ndarray): the constraint modal data.
            #. dataWeights (None, numpy.ndarray): the data points weights to
               compute the scale factor. If None is given, all data points
               will be considered as having the same weight.

        :Returns:
            #. scaleFactor (number): Constraint's scale factor or the
            new scale factor fit value.

        **NB**: This method WILL NOT UPDATE the internal scale factor
        value of the constraint.
        """
        SF = self.__scaleFactor
        # check to update scaleFactor
        if self.__adjustScaleFactorFrequency:
            if not self.engine.accepted%self.__adjustScaleFactorFrequency:
                SF = self.fit_scale_factor(experimentalData, modelData, dataWeights)
        return SF

    def compute_standard_error(self, experimentalData, modelData):
        """
        Compute the squared deviation between modal computed data
        and the experimental ones.

        .. math::
            SD = \\sum \\limits_{i}^{N} W_{i}(Y(X_{i})-F(X_{i}))^{2}

        Where:\n
        :math:`N` is the total number of experimental data points. \n
        :math:`W_{i}` is the data point weight. It becomes equivalent to 1
        when dataWeights is set to None. \n
        :math:`Y(X_{i})` is the experimental data point :math:`X_{i}`. \n
        :math:`F(X_{i})` is the computed from the model data  :math:`X_{i}`. \n

        :Parameters:
            #. experimentalData (numpy.ndarray): Experimental data.
            #. modelData (numpy.ndarray): The data to compare with the
               experimental one and compute the squared deviation.

        :Returns:
            #. standardError (number): The calculated standard error of
               the constraint.
        """
        # compute difference
        diff = experimentalData-modelData
        # return squared deviation
        if self.__dataWeights is None:
            return np.add.reduce((diff)**2)
        else:
            return np.add.reduce(self.__dataWeights*(diff)**2)

    def _get_needed_data_for_constraint_data_dictionary(self, *args, **kwargs):
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def get_constraint_data_dictionary(self, frame):
        """
        Get a dictionary of all constraint meaningful data to plot or export.

        :Parameters:
            #. frame (string): frame to pull and build contraint data. It can
               be a traditional frame, a multiframe or a subframe

        :Returns:
            #. data (dictionary): Dictionary of all meaningful data including:

                * output: list of frames dictionary constraint data
                * experimental_x: numpy array of experimental x data.
                * experimental_y: numpy array of experimental y data.
                * model_x: numpy array of model x data.
                * frames_weight: list of all frames weight.
                * frames_name: list of all frames name.
                * number_of_atoms_removed: list of number of removed atoms from
                  each frame.
                * shape_array: list of system shape function (numpy array) of
                  all frames.
                * window_function: list of window function (numpy array) of
                  all frames.
                * scale_factor: list of all frames scale factor.
                * standard_error: list of all frames standard error.
                * weighted_output: dictionary of all frames weighted constraint
                  data using 'frames_weight'
                * weighted_number_of_atoms_removed: All frames averaged number
                  of removed atoms using 'frames_weight'
                * weighted_scale_factor: All frames averaged scale factor
                  using 'frames_weight'
                * weighted_standard_error: All frames weighted standard error
                  using 'frames_weight'

        """
        from collections import OrderedDict
        neededData = self._get_needed_data_for_constraint_data_dictionary()
        isNormalFrame, isMultiframe, isSubframe = self.engine.get_frame_category(frame=frame)
        # get frames constraint data
        if isNormalFrame or isSubframe:
            frames       = [frame]
            framesName   = [frame,]
        else:
            frames       = [os.path.join(frame,frm) for frm in self.engine.frames[frame]['frames_name']]
            framesName   = self.engine.frames[frame]['frames_name']
        # initiate
        framesData = OrderedDict()
        framesData['output']                           = []
        framesData['experimental_x']                   = []
        framesData['experimental_y']                   = []
        framesData['model_x']                          = []
        framesData['frames_weight']                    = []
        framesData['frames_name']                      = []
        framesData['number_of_atoms_removed']          = []
        framesData['window_function']                  = []
        framesData['shape_array']                      = []
        framesData['scale_factor']                     = []
        framesData['standard_error']                   = []
        framesData['weighted_output']                  = None
        framesData['weighted_number_of_atoms_removed'] = None
        framesData['weighted_scale_factor']            = None
        framesData['weighted_standard_error']          = None
        # loop frames and get data
        _constraint = None
        for frm in frames:
            if frm == self.engine.usedFrame:
                output   = self._get_constraint_value(self.data, applyMultiframePrior=False)
                stdErr   = self.compute_standard_error(modelData = output["total"])
                expDis   = self._experimentalX
                expData  = self._experimentalY
                modelX   = self._modelX
                weight   = self.multiframeWeight if self.multiframeWeight is not None else FLOAT_TYPE(1.0)
                nRemAt   = len(self.engine._atomsCollector)
                sArray   = None
                if hasattr(self, '_shapeArray'):
                    sArray = self._shapeArray
                wFunc    = self.windowFunction
                sFactor  = self.scaleFactor
            else:
                repo = self.engine._get_repository()
                if _constraint is None:
                    _constraint = copy.deepcopy(self)
                object.__setattr__(_constraint.engine, '_Engine__usedFrame', frm)
                for name in neededData:
                    repoName = neededData[name]
                    if isinstance(name, tuple):
                        if name[0]=='engine':
                            value = repo.pull(relativePath=os.path.join(frm,repoName))
                            object.__setattr__(_constraint.engine, name[1], value)
                        else:
                            raise Exception(LOGGER.error('Wrong neededData format. Report issue'))
                    elif repoName.startswith('_Engine__'):
                        value = repo.pull(relativePath=os.path.join(frm,repoName))
                        object.__setattr__(_constraint.engine, name, value)
                    else:
                        value = repo.pull(relativePath=os.path.join(frm,'constraints',_constraint.constraintName,repoName))
                        object.__setattr__(_constraint, name, value)
                # append frames data
                if _constraint.data is None:
                    LOGGER.warn("%s constraint data for frame are not computed."%(self.__class__.__name__))
                    return None
                # get attributes
                output = _constraint._get_constraint_value(_constraint.data, applyMultiframePrior=False)
                stdErr = _constraint.compute_standard_error(modelData = output["total"])
                if len(framesData['experimental_x'])>0:
                    assert np.allclose(framesData['experimental_x'], _constraint._experimentalX), LOGGER.error("Frames experimentalX must match")
                    assert np.allclose(framesData['experimental_y'], _constraint._experimentalY), LOGGER.error("Frames experimentalY must match")
                    assert np.allclose(framesData['model_x'], _constraint._modelX), LOGGER.error("Frames modelX must match")
                else:
                    expDis   = _constraint._experimentalX
                    expData  = _constraint._experimentalY
                    modelX   = _constraint._modelX
                weight = _constraint.multiframeWeight if _constraint.multiframeWeight is not None else FLOAT_TYPE(1.0)
                nRemAt   = len(_constraint.engine._atomsCollector)
                sArray   = None
                if hasattr(self, '_shapeArray'):
                    sArray = _constraint._shapeArray
                wFunc    = _constraint.windowFunction
                sFactor  = _constraint.scaleFactor
            # append frame data
            framesData['output'].append(output)
            framesData['frames_name'].append(frm)
            framesData['experimental_x'].append(expDis)
            framesData['experimental_y'].append(expData)
            framesData['model_x'].append(modelX)
            framesData['frames_weight'].append(weight)
            framesData['number_of_atoms_removed'].append(nRemAt)
            framesData['shape_array'].append(sArray)
            framesData['window_function'].append(wFunc)
            framesData['scale_factor'].append(sFactor)
            framesData['standard_error'].append(stdErr)
        ## compute frames weighted
        framesWeight = np.array(framesData['frames_weight'], dtype=FLOAT_TYPE)
        if np.sum(framesWeight)==0:
            if len(framesWeight)==1:
                framesWeight = np.array([1])
            else:
                raise Exception("Frames weight sum is found to be 0. PLEASE REPORT THIS BUG")
        weightedOutput = {}
        for key in framesData['output'][0]:
            if any([framesData['output'][i][key] is None for i, w in enumerate(framesWeight)]):
                weightedOutput[key] = None
            else:
                weightedOutput[key] = np.sum([w*framesData['output'][i][key] for i, w in enumerate(framesWeight)], axis=0)
                weightedOutput[key] = weightedOutput[key]/sum(framesWeight)
        framesData['weighted_output']                  = weightedOutput
        framesData['weighted_number_of_atoms_removed'] = np.average(framesData['number_of_atoms_removed'], weights=framesWeight)
        framesData['weighted_standard_error']          = self.compute_standard_error(modelData=weightedOutput['total'])
        framesData['weighted_scale_factor']            = np.average(framesData['scale_factor'], weights=framesWeight)
        # return
        return framesData


    def _plot(self, frame, output, experimentalX, experimentalY,
                    shellCenters, shapeArray, numberOfRemovedAtoms,
                    standardError, scaleFactor, multiframeWeight,
                    # plotting arguments
                    ax=None, intra=True, inter=True, totalNoWindow=False,
                    xlabelParams=None,
                    ylabelParams=None,
                    legendParams=None,
                    titleFormat = "",
                    expParams = {'label':"experimental","color":'red','marker':'o','markersize':7.5, 'markevery':1, 'zorder':0},
                    totParams = {'label':"total", 'color':'black','linewidth':3.0, 'zorder':1},
                    noWParams = {'label':"total - no window", 'color':'black','linewidth':1.0, 'zorder':1},
                    shaParams = {'label':"shape function", 'color':'black','linewidth':1.0, 'linestyle':'dashed'},
                    parParams = {'linewidth':1.0, 'markevery':5, 'markersize':5, 'zorder':-1},
                    show=True):
        # import matplotlib
        import matplotlib.pyplot as plt
        # get axes
        if ax is None:
            FIG  = plt.figure()
            AXES = plt.gca()
        else:
            AXES = ax
            FIG  = AXES.get_figure()
        # Create plotting styles
        COLORS  = ["b",'g','c','y','m']
        MARKERS = ["",'.','+','^','|']
        INTRA_STYLES = [r[0] + r[1]for r in itertools.product(['--'], list(reversed(COLORS)))]
        INTRA_STYLES = [r[0] + r[1]for r in itertools.product(MARKERS, INTRA_STYLES)]
        INTER_STYLES = [r[0] + r[1]for r in itertools.product(['-'], COLORS)]
        INTER_STYLES = [r[0] + r[1]for r in itertools.product(MARKERS, INTER_STYLES)]
        # plot experimental
        AXES.plot(experimentalX,experimentalY, **expParams)
        AXES.plot(shellCenters, output["total"], **totParams )
        if totalNoWindow and output["total_no_window"] is not None:
            AXES.plot(shellCenters, output["total_no_window"], **noWParams )
        if shapeArray is not None:
            AXES.plot(shellCenters, shapeArray, **shaParams )
        # plot partials
        intraStyleIndex = 0
        interStyleIndex = 0
        for key in output:
            val = output[key]
            if key in ("total", "total_no_window"):
                continue
            elif "intra" in key and intra:
                AXES.plot(shellCenters, val, INTRA_STYLES[intraStyleIndex], label=key, **parParams )
                intraStyleIndex+=1
            elif "inter" in key and inter:
                AXES.plot(shellCenters, val, INTER_STYLES[interStyleIndex], label=key, **parParams )
                interStyleIndex+=1
        # plot legend
        if legendParams is not None:
            AXES.legend(**legendParams)
        # set title
        if len(titleFormat):
            name = ' '.join(re.findall('[A-Z][^A-Z]*', self.__class__.__name__))
            FIG.canvas.set_window_title(name)
            # format title
            title = titleFormat.format(frame=frame,
                                       standardError=standardError,
                                       numberOfRemovedAtoms=numberOfRemovedAtoms,
                                       scaleFactor=scaleFactor,
                                       multiframeWeight=multiframeWeight)
            AXES.set_title(title)
        # set axis labels
        if xlabelParams is not None:
            #AXES.set_xlabel("$r(\AA)$", size=16)
            AXES.set_xlabel(**xlabelParams)
        if ylabelParams is not None:
            AXES.set_ylabel(**ylabelParams)
        # set background color
        FIG.patch.set_facecolor('white')
        #show
        if show:
            plt.show()
        return FIG, AXES

    def plot(self, frame=None, multiplot=False,
                   ax=None, intra=True, inter=True, shapeFunc=True,
                   xlabelParams=True,
                   ylabelParams=True,
                   legendParams=True,
                   expParams=None,
                   totParams=None,
                   noWParams=None,
                   shaParams=None,
                   parParams=None,
                   titleFormat = "@{frame} (${numberOfRemovedAtoms:.1f}$ $rem.$ $at.$) $Std.Err.={standardError:.3f}$\n$scale$ $factor$=${scaleFactor:.2f}$ - $multiframe$ $weight$=${multiframeWeight:.3f}$",
                   show=True):
        """
        Plot constraint data

        :Parameters:
            #. frame (None, string): The frame name to plot. If None, used frame
               will be plotted.
            #. multiplot (boolean): If multiframe is given, multiplot insures
               plotting all frames differently in a single multiplot figure.
               In this case, given ax is ommitted.
            #. ax (None, matplotlib Axes): matplotlib Axes instance to plot in.
               If None is given a new plot figure will be created.
            #. intra (boolean): Whether to add intra-molecular pair
               distribution function features to the plot.
            #. inter (boolean): Whether to add inter-molecular pair
               distribution function features to the plot.
            #. shapeFunc (boolean): Whether to add shape function to the plot
               only when exists.
            #. xlabelParams (dict, boolean): matplotlib.axes.Axes.set_xlabel
               parameters. If True, default parameters given by constraint's
               property _plotDefaultParameters is set. If False x axes label is
               omitted
            #. ylabelParams (dict, boolean): matplotlib.axes.Axes.set_ylabel
               parameters. If True,  default parameters given by constraint's
               property _plotDefaultParameters is set. If False y axes label is
               omitted
            #. legendParams (dict, boolean):matplotlib.axes.Axes.legend
               parameters. If True,  default parameters given by constraint's
               property _plotDefaultParameters is set. If False is omitted
            #. titleFormat (string): title format. If empty string is given no
               title will be added to figure axes
            #. expParams (None, dict): experimental data matplotlib.pyplot.plot
               parameters. If None,  default parameters given by constraint's
               property _plotDefaultParameters is set. If dict, it will be the
               update dict to _plotDefaultParameters
            #. totParams (None, dict): total data matplotlib.pyplot.plot
               parameters. If None,  default parameters given by constraint's
               property _plotDefaultParameters is set. If dict, it will be the
               update dict to _plotDefaultParameters
            #. noWParams (None, dict): constraint no-window total data matplotlib.pyplot.plot
               parameters. If None,  default parameters given by constraint's
               property _plotDefaultParameters is set. If dict, it will be the
               update dict to _plotDefaultParameters
            #. shaParams (None, dict): constraint shape array data matplotlib.pyplot.plot
               parameters. If None,  default parameters given by constraint's
               property _plotDefaultParameters is set. If dict, it will be the
               update dict to _plotDefaultParameters
            #. parParams (None, dict): constraint partial data matplotlib.pyplot.plot
               parameters. If None,  default parameters given by constraint's
               property _plotDefaultParameters is set. If dict, it will be the
               update dict to _plotDefaultParameters
            #. show (boolean): Whether to render and show figure before
               returning.

        :Returns:
            #. figure (matplotlib Figure): matplotlib used figure.
            #. axes (matplotlib Axes): matplotlib used axes.
        """
        assert isinstance(multiplot, bool), LOGGER.error("multiplot must be boolean")
        # check expParams
        if expParams is None:
            expParams = {}
        assert isinstance(expParams, dict), "expParams must be None or dict"
        default   = copy.deepcopy( self._plotDefaultParameters['expParams'] )
        default.update(expParams)
        expParams = default
        # check totParams
        if totParams is None:
            totParams = {}
        assert isinstance(totParams, dict), "totParams must be None or dict"
        default   = copy.deepcopy( self._plotDefaultParameters['totParams'] )
        default.update(totParams)
        totParams = default
        # check noWParams
        if noWParams is None:
            noWParams = {}
        assert isinstance(noWParams, dict), "noWParams must be None or dict"
        default   = copy.deepcopy( self._plotDefaultParameters['noWParams'] )
        default.update(noWParams)
        noWParams = default
        # check shaParams
        if shaParams is None:
            shaParams = {}
        assert isinstance(shaParams, dict), "shaParams must be None or dict"
        default   = copy.deepcopy( self._plotDefaultParameters['shaParams'] )
        default.update(shaParams)
        shaParams = default
        # check parParams
        if parParams is None:
            parParams = {}
        assert isinstance(parParams, dict), "parParams must be None or dict"
        default   = copy.deepcopy( self._plotDefaultParameters['parParams'] )
        default.update(parParams)
        parParams = default
        # check xlabel
        if xlabelParams is False:
            xlabelParams = None
        elif xlabelParams is True:
            xlabelParams = self._plotDefaultParameters['xlabelParams']
        elif  isinstance(xlabelParams, dict):
            default   = copy.deepcopy( self._plotDefaultParameters['xlabelParams'] )
            default.update(xlabelParams)
            xlabelParams = default
        else:
            assert xlabelParams is None, "xlabelParams must be None, boolean or dict"
        # check ylabel
        if ylabelParams is False:
            ylabelParams = None
        elif ylabelParams is True:
            ylabelParams = self._plotDefaultParameters['ylabelParams']
        elif  isinstance(ylabelParams, dict):
            default = copy.deepcopy( self._plotDefaultParameters['ylabelParams'] )
            default.update(ylabelParams)
            ylabelParams = default
        else:
            assert ylabelParams is None, "ylabelParams must be None, boolean or dict"
        # check ylabel
        if legendParams is False:
            legendParams = None
        elif legendParams is True:
            legendParams = self._plotDefaultParameters['legendParams']
        elif  isinstance(legendParams, dict):
            default = copy.deepcopy( self._plotDefaultParameters['legendParams'] )
            default.update(legendParams)
            legendParams = default
        else:
            assert legendParams is None, "legendParams must be None, boolean or dict"
        # get frame
        if frame is None:
            frame = self.engine.usedFrame
        isNormalFrame, isMultiframe, isSubframe = self.engine.get_frame_category(frame=frame)
        framesData = self.get_constraint_data_dictionary(frame=frame)
        if framesData is None:
            return
        # get number of frames
        numberOfFrames = len(framesData['frames_weight'])
        multiplot      = multiplot and numberOfFrames>1
        if multiplot:
            import matplotlib.pyplot as plt
            nrows     = int(np.sqrt(float(numberOfFrames)))
            ncols     = int(np.ceil(float(numberOfFrames)/nrows))
            fig, axes = plt.subplots(nrows=nrows, ncols=ncols)
            axes      = axes.flatten()
            fig.subplots_adjust(hspace=0.4)
            for idx, output in enumerate(framesData['output']):
                _show = show and idx==len(framesData['output'])-1
                FIG, AXES = self._plot(frame                 = framesData['frames_name'][idx],
                                       output                = output,
                                       experimentalX         = framesData['experimental_x'][idx],
                                       experimentalY         = framesData['experimental_y'][idx],
                                       shellCenters          = framesData['model_x'][idx],
                                       shapeArray            = framesData['shape_array'][idx] if shapeFunc else None,
                                       numberOfRemovedAtoms  = framesData['number_of_atoms_removed'][idx],
                                       standardError         = framesData['standard_error'][idx],
                                       scaleFactor           = framesData['scale_factor'][idx],
                                       multiframeWeight      = framesData['frames_weight'][idx],
                                       ax=axes[idx], intra=intra, inter=inter,
                                       xlabelParams = xlabelParams,
                                       ylabelParams = ylabelParams,
                                       legendParams = legendParams,
                                       expParams    = expParams,
                                       totParams    = totParams,
                                       noWParams    = noWParams,
                                       shaParams    = shaParams,
                                       parParams    = parParams,
                                       titleFormat  = titleFormat,
                                       show=_show)
        else:
            FIG, AXES = self._plot(frame                 = frame,
                                   output                = framesData['weighted_output'],
                                   experimentalX         = framesData['experimental_x'][0],
                                   experimentalY         = framesData['experimental_y'][0],
                                   shellCenters          = framesData['model_x'][0],
                                   shapeArray            = framesData['shape_array'][0] if shapeFunc else None,
                                   numberOfRemovedAtoms  = framesData['weighted_number_of_atoms_removed'],
                                   standardError         = framesData['weighted_standard_error'],
                                   scaleFactor           = framesData['weighted_scale_factor'],
                                   multiframeWeight      = np.sum(framesData['frames_weight']),
                                   ax=ax, intra=intra, inter=inter,
                                   xlabelParams = xlabelParams,
                                   ylabelParams = ylabelParams,
                                   legendParams = legendParams,
                                   expParams    = expParams,
                                   totParams    = totParams,
                                   noWParams    = noWParams,
                                   shaParams    = shaParams,
                                   parParams    = parParams,
                                   titleFormat  = titleFormat,
                                   show=show)
        # return
        return FIG, AXES


    def plot_multiframe_weights(self, frame, ax=None, titleFormat = "@{frame} [subframes probability distribution]", show=True):
        """ plot multiframe subframes weight distribution histogram

        :Parameters:
           #. frame (string): multiframe name
           #. ax (None, matplotlib Axes): matplotlib Axes instance to plot in.
              If None is given a new plot figure will be created.
           #. titleFormat (string): title format. If empty string is given no
              title will be added to figure axes
           #. show (boolean): Whether to render and show figure before
               returning.

        :Returns:
            #. figure (matplotlib Figure): matplotlib used figure.
            #. axes (matplotlib Axes): matplotlib used axes.

        """
        # import matplotlib
        import matplotlib.pyplot as plt
        # check frame
        if frame is None:
            frame = self.engine.usedFrame
        isNormalFrame, isMultiframe, isSubframe = self.engine.get_frame_category(frame=frame)
        assert isMultiframe, "Given frame '%s' is not a multiframe"%frame
        # get axes
        if ax is None:
            FIG  = plt.figure()
            AXES = plt.gca()
        else:
            AXES = ax
            FIG  = AXES.get_figure()
        #
        data   = self.get_constraint_data_dictionary(frame=frame)
        totals = [i['total'] for i in data['output']]
        frames = data['frames_name']
        stderr = data['standard_error']
        ratios = data['frames_weight']
        # plot bars
        bars = plt.bar(range(len(frames)), ratios, align='center', alpha=0.5)
        #plt.xticks(range(len(frames)), frames, rotation=90)
        plt.xticks(range(len(frames)), self.engine.frames[frame]['frames_name'])
        # plot weights text
        pos  = plt.gca().get_xticks()
        ylim = plt.gca().get_ylim()
        diff = 0.05*abs(ylim[1] - ylim[0])
        for i in range(len(ratios)):
            y = ratios[i]+diff
            if y < 0.35*ylim[1]:
                y = 0.35*ylim[1]
            if y >  0.65*ylim[1]:
                y = 0.65*ylim[1]
            plt.text(x=pos[i] , y=max(0.35*ylim[1], ratios[i]+diff), s='%.4f (%.2f)'%(ratios[i],stderr[i]), size = 10, ha='center', va='top', rotation=90)
        # set labels
        plt.ylabel('Coefficient')
        plt.xlabel('Sub-frames')
        # set title
        if len(titleFormat):
            name = ' '.join(re.findall('[A-Z][^A-Z]*', self.__class__.__name__))
            FIG.canvas.set_window_title(name)
            # format title
            title = titleFormat.format(frame=frame)
            AXES.set_title(title)
        # show
        if show:
            plt.show()
        # return
        return FIG, AXES

    def _export(self, fileName, framesData, format='%12.5f', delimiter='\t', comments='# '):
        # create data, metadata and header
        metadata = []
        header   = []
        data     = []
        if 'frames_name' in framesData:
            metadata = ["Frames name:                    %s"%framesData['frames_name'],
                        "frames number of removed atoms: %s"%dict(list(zip(framesData['frames_name'],framesData['number_of_atoms_removed']))),
                        "frames multiframe weights:      %s"%dict(list(zip(framesData['frames_name'],framesData['frames_weight']))),
                        "Frames scale factor:            %s"%dict(list(zip(framesData['frames_name'],framesData['scale_factor']))),
                        "Frames standard error:          %s"%dict(list(zip(framesData['frames_name'],framesData['standard_error']))),
                        ]
            for idx, frm in enumerate(framesData['frames_name']):
                # append experimental distances
                header.append('%s:experimental_x'%frm)
                data.append(framesData['experimental_x'][idx])
                # append experimental data
                header.append('%s:experimental_y'%frm)
                data.append(framesData['experimental_y'][idx])
                # append experimental data
                header.append('%s:model_x'%frm)
                data.append(framesData['model_x'][idx] )
                # loop all outputs
                output = framesData['output'][idx]
                for dn in output:
                    header.append('%s:%s'%(frm,dn))
                    data.append(output[dn])
        # append weighted data
        if len([k for k in framesData if k.startswith('weighted_')]):
            metadata.append("Weighted number of removed atoms: %s"%framesData['weighted_number_of_atoms_removed'])
            metadata.append("Weighted frames scale factor:     %s"%framesData['weighted_scale_factor'])
            metadata.append("Weighted frames standard error:   %s"%framesData['weighted_standard_error'])
            output = framesData['weighted_output']
            for dn in output:
                header.append('weighted_%s'%(dn))
                data.append(output[dn])
        # finalize metadata
        metadata.append(" ".join(header))
        # create array and export
        data = np.transpose(data).astype(float)
        # save
        np.savetxt(fname     = fileName,
                   X         = data,
                   fmt       = format,
                   delimiter = delimiter,
                   header    = "\n".join(metadata),
                   comments  = comments)


    def export(self, fileName, frame=None, splitFrames=False, format='%12.5f', delimiter=' ', comments='# '):
        """
        Export constraint data to text file or to an archive of files.

        :Parameters:
            #. fileName (path): full file name and path.
            #. frame (None, string): frame name to export data from. If multiframe
               is given, multiple files will be created with subframe name
               appended to the end.
            #. splitFrames (boolean): whether to split frames into multiple
               files and create and archive (.zip) of frames data. If given
               frame is a normal frame or a subframe this option is not used.
            #. format (string): string format to export the data.
               format is as follows (%[flag]width[.precision]specifier)
            #. delimiter (string): String or character separating columns.
            #. comments (string): String that will be prepended to the header.
        """
        if frame is None:
            frame = self.engine.usedFrame
        framesData = self.get_constraint_data_dictionary(frame=frame)
        if framesData is None:
            return
        # get number of frames
        if len(framesData['frames_name'])==1 or not splitFrames:
            self._export(fileName = fileName,
                         framesData = framesData,
                         format=format, delimiter=delimiter,
                         comments=comments )
        else:
            dirname, filename   = os.path.split(fileName)
            if not len(dirname):
                dirname = os.getcwd()
            filename, extension = os.path.splitext(filename)
            zipFilePath = os.path.join(dirname, filename)
            if os.path.isfile( zipFilePath+'.zip' ):
                LOGGER.warn("File '%s' is removed and replaced with new one"%(zipFilePath,))
            _dirname = os.path.join(dirname, '.'+str(uuid.uuid1()))
            os.makedirs(_dirname)
            try:
                for idx, frameName in enumerate(framesData['frames_name']):
                    _framesData = dict([(key,[framesData[key][idx]]) for key in framesData if not key.startswith('weighted_')])
                    self._export(framesData=_framesData,
                                 fileName=os.path.join(_dirname,'%s.txt'%frameName.replace(os.sep, '_')),
                                 format=format, delimiter=delimiter,
                                 comments=comments )
                # exported weighted
                _framesData = dict([(key,framesData[key]) for key in framesData if key.startswith('weighted_')])
                self._export(framesData=_framesData,
                             fileName=os.path.join(_dirname,'%s_weighted.txt'%frame.replace(os.sep, '_')),
                             format=format, delimiter=delimiter,
                             comments=comments )
                # create zip file
                shutil.make_archive(zipFilePath, 'zip', _dirname)
            except Exception as err:
                print(Exception(err))
            finally:
                # remove directory
                shutil.rmtree(_dirname)



class SingularConstraint(Constraint):
    """ A singular constraint is a constraint that doesn't allow multiple
    instances in the same engine."""

    def is_singular(self, engine):
        """
        Get whether only one instance of this constraint type is present
        in the stochastic engine. True for only itself found, False for
        other instance of the same __class__.__name__ or constraintId.

        :Parameters:
            #. engine (stochastic fullrmc engine): Engine instance.

        :Returns:
            #. result (boolean): Whether constraint is singular in engine.
        """
        for c in engine.constraints:
            if c is self or c.constraintId==self.constraintId:
                continue
            if c.__class__.__name__ == self.__class__.__name__:
                return False
        return True

    def assert_singular(self, engine):
        """
        Checks whether only one instance of this constraint type is
        present in the stochastic engine. Raises Exception if multiple
        instances are present.
        """
        assert self.is_singular(engine), LOGGER.error("Only one instance of constraint '%s' is allowed in the same engine"%self.__class__.__name__)


class RigidConstraint(Constraint):
    """
    A rigid constraint is a constraint that doesn't count into the total
    standard error of the stochastic Engine. But it's internal standard error
    must monotonously decrease or remain the same from one engine step to
    another. If standard error of an rigid constraint increases the
    step will be rejected even before engine's new standardError get computed.

    :Parameters:
        #. rejectProbability (Number): Rejecting probability of all steps
           where standard error increases. It must be between 0 and 1 where
           1 means rejecting all steps where standardError increases
           and 0 means accepting all steps regardless whether standard error
           increases or not.
    """
    def __init__(self, rejectProbability):
        # initialize constraint
        super(RigidConstraint, self).__init__()
        # set probability
        self.set_reject_probability(rejectProbability)
        # set frame data
        FRAME_DATA = [d for d in self.FRAME_DATA]
        FRAME_DATA.extend(['_RigidConstraint__rejectProbability'] )
        object.__setattr__(self, 'FRAME_DATA',  tuple(FRAME_DATA) )

    @property
    def rejectProbability(self):
        """ Rejection probability. """
        return self.__rejectProbability

    def set_reject_probability(self, rejectProbability):
        """
        Set the rejection probability. This method doesn't allow specifying
        frames. It will target used frame only.

        :Parameters:
            #. rejectProbability (Number): rejecting probability of all steps
               where standard error increases. It must be between 0 and 1
               where 1 means rejecting all steps where standardError increases
               and 0 means accepting all steps regardless whether standard
               error increases or not.
        """
        assert is_number(rejectProbability), LOGGER.error("rejectProbability must be a number")
        rejectProbability = FLOAT_TYPE(rejectProbability)
        assert rejectProbability>=0 and rejectProbability<=1, LOGGER.error("rejectProbability must be between 0 and 1")
        self.__rejectProbability = rejectProbability
        # dump to repository
        self._dump_to_repository({'_RigidConstraint__rejectProbability': self.__rejectProbability})

    def should_step_get_rejected(self, standardError):
        """
        Given a standard error, return whether to keep or reject new
        standard error according to the constraint reject probability.

        :Parameters:
            #. standardError (number): The standard error to compare with
            the Constraint standard error

        :Return:
            #. result (boolean): True to reject step, False to accept
        """
        if self.standardError is None:
            raise Exception(LOGGER.error("must compute data first"))
        if standardError<=self.standardError:
            return False
        return randfloat() < self.__rejectProbability

    def should_step_get_accepted(self, standardError):
        """
        Given a standard error, return whether to keep or reject new standard
        error according to the constraint reject probability.

        :Parameters:
            #. standardError (number): The standard error to compare with
               the Constraint standard error

        :Return:
            #. result (boolean): True to accept step, False to reject
        """
        return not self.should_step_get_reject(standardError)


#class QuasiRigidConstraint(RigidConstraint):
#    """
#    A quasi-rigid constraint is a another rigid constraint but it becomes free
#    above a certain threshold ratio of satisfied data. Every quasi-rigid
#    constraint has its own definition of maximum standard error. The ratio is
#    computed as between current standard error and maximum standard error.
#
#     .. math::
#
#        ratio = 1-\\frac{current\ standard\ error}{maximum\ standard\ error}
#
#    :Parameters:
#        #. rejectProbability (Number): Rejecting probability of all steps
#           where standardError increases only before threshold ratio is reached.
#           It must be between 0 and 1 where 1 means rejecting all steps where
#           standardError increases and 0 means accepting all steps regardless
#           whether standardError increases or not.
#        #. thresholdRatio(Number): The threshold of satisfied data, above
#           which the constraint become free. It must be between 0 and 1 where
#           1 means all data must be satisfied and therefore the constraint
#           behave like a RigidConstraint and 0 means none of the data must be
#           satisfied and therefore the constraint becomes always free and
#           useless.
#    """
#    def __init__(self, engine, rejectProbability, thresholdRatio):
#        # initialize constraint
#        super(QuasiRigidConstraint, self).__init__(rejectProbability=rejectProbability)
#        # set probability
#        self.set_threshold_ratio(thresholdRatio)
#        # initialize maximum standard error
#        #self.__maximumStandardError = None
#        self._set_maximum_standard_error(None)
#        # set frame data
#        FRAME_DATA = [d for d in self.FRAME_DATA]
#        FRAME_DATA.extend(['_QuasiRigidConstraint__thresholdRatio',
#                           '_QuasiRigidConstraint__maximumStandardError'] )
#        object.__setattr__(self, 'FRAME_DATA',  tuple(FRAME_DATA) )
#
#
#    def _set_maximum_standard_error(self, maximumStandardError):
#        """ Set the maximum standard error. Use carefully, it's not meant to
#        be used externally. maximum squared deviation is what is used to
#        compute the ratio and compare to threshold ratio.
#        """
#        if maximumStandardError is not None:
#            assert is_number(maximumStandardError), LOGGER.error("maximumStandardError must be a number.")
#            maximumStandardError = FLOAT_TYPE(maximumStandardError)
#            assert maximumStandardError>0, LOGGER.error("maximumStandardError must be a positive.")
#        self.__maximumStandardError = maximumStandardError
#        # dump to repository
#        self._dump_to_repository({'_QuasiRigidConstraint__maximumStandardError': self.__maximumStandardError})
#
#    @property
#    def thresholdRatio(self):
#        """ Threshold ratio. """
#        return self.__thresholdRatio
#
#    @property
#    def currentRatio(self):
#        return 1-(self.standardError/self.__maximumStandardError)
#
#    def set_threshold_ratio(self, thresholdRatio):
#        """
#        Set the rejection probability function.
#
#        :Parameters:
#            #. thresholdRatio(Number): The threshold of satisfied data, above
#               which the constraint become free. It must be between 0 and 1
#               where 1 means all data must be satisfied and therefore the
#               constraint behave like a RigidConstraint and 0 means none of
#               the data must be satisfied and therefore the constraint
#               becomes always free and useless.
#        """
#        assert is_number(thresholdRatio), LOGGER.error("thresholdRatio must be a number")
#        thresholdRatio = FLOAT_TYPE(thresholdRatio)
#        assert thresholdRatio>=0 and thresholdRatio<=1, LOGGER.error("thresholdRatio must be between 0 and 1")
#        self.__thresholdRatio = thresholdRatio
#        # dump to repository
#        self._dump_to_repository({'_QuasiRigidConstraint__thresholdRatio': self.__thresholdRatio})
#
#    def should_step_get_rejected(self, standardError):
#        """
#        Given a standard error, return whether to keep or reject new
#        standard error according to the constraint reject probability function.
#
#        :Parameters:
#            #. standardError (number): Standard error to compare with the
#            Constraint standard error.
#
#        :Return:
#            #. result (boolean): True to reject step, False to accept.
#        """
#        previousRatio = 1-(self.standardError/self.__maximumStandardError)
#        currentRatio  = 1-(standardError/self.__maximumStandardError)
#        if currentRatio>=self.__thresholdRatio: # must be accepted
#            return False
#        elif previousRatio>=self.__thresholdRatio: # it must be rejected
#            return randfloat() < self.rejectProbability
#        elif standardError<=self.standardError: # must be accepted
#            return False
#        else: # must be rejected
#            return randfloat() < self.rejectProbability
#



"""
In order to work properly, this script must be put one layer or directory
outside of fullrmc package directory.
"""
##########################################################################################
########################################  IMPORTS  #######################################
# standard distribution imports
from __future__ import print_function
import os, sys, subprocess
import fnmatch

# setup and distutils imports
try:
    from setuptools import setup
except:
    from distutils.core import setup
from distutils.util import convert_path
from distutils.core import Extension

# import cython
try:
    from Cython.Distutils import build_ext
except:
    raise Exception("must install cython first. Try pip install cython")

# import numpy
try:
    import numpy as np
except:
    raise Exception("must install numpy first. Try pip install numpy")


# check python version
if sys.version_info.major == 2:
    assert sys.version_info.minor>=7, "Only ython 2.7.y is accepted in the 2.x.y version of python"
else:
    assert sys.version_info.major == 3, "Only python 2.7.y or python 3.7.y and above are accepted versions"
    #assert sys.version_info.minor>=7, "Only python 3.7.y and above is accepted in the 3.x.y version of python"

## get python 2 and 3 compatibilities
if sys.version_info.major==3:
    # THIS IS PYTHON 3
    str        = str
    long       = int
    unicode    = str
    bytes      = bytes
    basestring = str
    xrange     = range
    range      = lambda *args: list( xrange(*args) )
    maxint     = sys.maxsize
else:
    # THIS IS PYTHON 2
    str        = str
    unicode    = unicode
    bytes      = str
    long       = long
    basestring = basestring
    xrange     = xrange
    range      = range
    maxint     = sys.maxint


##########################################################################################
#################################  PACKAGE PATH AND NAME  ################################
PACKAGE_PATH    = '.'
PACKAGE_NAME    = 'fullrmc'
EXTENSIONS_PATH = os.path.join(PACKAGE_NAME, "Extensions")



##########################################################################################
######################################  MANIFEST.in  #####################################
commands = [# include MANIFEST.in
            '# include this file, to ensure we can recreate source distributions',
            'include MANIFEST.in'
            # exclude all .log files
            '\n# exclude all logs',
            'global-exclude *.log',
            # exclude all .log files
            '\n# exclude all c cpp and compiled files',
            'global-exclude *.c',
            'global-exclude *.cpp',
            'global-exclude *.so',
            'global-exclude *.rmc',
            # exclude all other non necessary files
            '\n# exclude all other non necessary files ',
            'global-exclude .project',
            'global-exclude .pydevproject',
            # exclude all of the subversion metadata
            '\n# exclude all of the subversion metadata',
            'global-exclude *.svn*',
            'global-exclude .svn/*',
            'global-exclude *.git*',
            'global-exclude .git/*',
            # exclude all
            'global-exclude %s/Examples/*.dat'%PACKAGE_NAME,
            # include all Example files
            '\n# include all Example files',
            'global-include %s/Examples/*.py'%PACKAGE_NAME,
            'global-include %s/Examples/*/*.py'%PACKAGE_NAME,
            'global-include %s/Examples/*/*.exp'%PACKAGE_NAME,
            'global-include %s/Examples/*/*.pdb'%PACKAGE_NAME,
            # include all LICENCE files
            '\n# include all license files found',
            'global-include %s/*LICENSE.*'%PACKAGE_NAME,
            # include all README files
            '\n# include all readme files found',
            'global-include %s/*README.*'%PACKAGE_NAME,
            'global-include %s/*readme.*'%PACKAGE_NAME,
            # exclude unnecessary files
            'global-exclude %s/Examples/*/*.dat'%PACKAGE_NAME,
            'global-exclude %s/Examples/*/visualizeTrajectory.py'%PACKAGE_NAME,
            'global-exclude %s/Examples/*/plotFigures.py'%PACKAGE_NAME,
            'global-exclude %s/Examples/*/extractTrajectory.py'%PACKAGE_NAME,
            'global-exclude %s/Examples/*/exportTrajectoryFigures.py'%PACKAGE_NAME,
            'global-exclude %s/Examples/*/extractTrajectoryPDF.py'%PACKAGE_NAME,
            'global-exclude %s/Examples/*/restart.pdb'%PACKAGE_NAME,
            ]


with open('MANIFEST.in','w') as fd:
    for l in commands:
        fd.write(l)
        fd.write('\n')



##########################################################################################
######################################  CLASSIFIERS  #####################################
CLASSIFIERS = """\
Development Status :: 4 - Beta
Intended Audience :: Science/Research
Intended Audience :: Education
Intended Audience :: Developers
Natural Language :: English
License :: OSI Approved :: GNU Affero General Public License v3
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.7
Programming Language :: Cython
Topic :: Software Development
Topic :: Software Development :: Build Tools
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Artificial Intelligence
Topic :: Scientific/Engineering :: Physics
Topic :: Scientific/Engineering :: Mathematics
Topic :: Scientific/Engineering :: Chemistry
Operating System :: OS Independent
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""



##########################################################################################
######################################  DESCRIPION  ######################################
# create descriptions
LONG_DESCRIPTION = ["FUndamental Library Language for Reverse Monte Carlo or fullrmc is a molecular/atomic stochastic fitting platform to reverse modeling experimental data. ",
                    "fullrmc is not a standard RMC software but exceeds in its capabilities and functionalities traditional RMC and Metropolis-Hastings algoritm. ",
                    "Therefore RMC appellation in fullrmc, is not accurate but it's retained to respecting the community terminology. ",
                    "RMC is probably best known for its applications in condensed matter physics and solid state chemistry. ",
                    "RMC is used to solve an inverse problem whereby an atomic model is adjusted until its atoms position have the greatest consistency with a set of experimental data.",
                    "fullrmc is a python package with its core and calculation modules optimized and compiled in Cython. ",
                    "fullrmc's Engine sub-module is the main module that contains the definition of 'Engine' which is the main and only class used to launch the stochastic calculation. ",
                    "Engine reads only Protein Data Bank formatted atomic configuration files '.pdb' and handles other definitions and attributes. ",
                    "Starting from version 1.x.y fitting non-periodic boundary conditions or isolated molecules is added. ",
                    "fullrmc >= 1.2.y can be compiled with 'openmp' allowing multithreaded fitting. ",
                    "fullrmc >= 2.x.y engine is a single file no more but a pyrep repository. ",
                    "fullrmc >= 3.x.y dynamically removing atoms upon fitting is enabled. ",]
DESCRIPTION      = [ LONG_DESCRIPTION[0] ]
DESCRIPTION      = [ LONG_DESCRIPTION[0] ]

# get package info
PACKAGE_INFO={}
if sys.version_info.major == 2:
    execfile(convert_path( os.path.join(PACKAGE_PATH, PACKAGE_NAME,'__pkginfo__.py') ), PACKAGE_INFO)
else:
    exec(open(convert_path( os.path.join(PACKAGE_PATH, PACKAGE_NAME,'__pkginfo__.py') )).read(), PACKAGE_INFO)


##########################################################################################
################################### USEFUL DEFINITIONS ###################################
def is_package(path):
    return (os.path.isdir(path) and os.path.isfile(os.path.join(path, '__init__.py')))

def get_packages(path, base="", exclude=None):
    if exclude is None:
        exclude = []
    assert isinstance(exclude, (list, set, tuple)), "exclude must be a list"
    exclude = [os.path.abspath(e) for e in exclude]
    packages = {}
    for item in os.listdir(path):
        d = os.path.join(path, item)
        if sum([e in os.path.abspath(d) for e in exclude]):
            continue
        if is_package(d):
            if base:
                module_name = "%(base)s.%(item)s" % vars()
            else:
                module_name = item
            packages[module_name] = d
            packages.update(get_packages(d, module_name, exclude))
    return packages

DATA_EXCLUDE = ('*.rmc','*.data','*.png','*.xyz','*.log','*.pyc', '*~', '.*', '*.so', '*.pyd')
EXCLUDE_DIRECTORIES = ('*svn','*git','dist', 'EGG-INFO', '*.egg-info',"*.rmc")
def find_package_data(where='.', package='', relativePath='',
                      exclude=DATA_EXCLUDE, excludeDirectories=EXCLUDE_DIRECTORIES,
                      onlyInPackages=True, showIgnored=False):
    out = {}
    stack = [(convert_path(where), '', package, onlyInPackages)]
    while stack:
        where, prefix, package, onlyInPackages = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if os.path.isdir(fn):
                excludeThis = False
                for pattern in excludeDirectories:
                    if (fnmatch.fnmatchcase(name, pattern) or fn.lower() == pattern.lower()):
                        excludeThis = True
                        if showIgnored:
                            print >> sys.stderr, ("Directory %s ignored by pattern %s" % (fn, pattern))
                        break
                if excludeThis:
                    continue
                if (os.path.isfile(os.path.join(fn, '__init__.py')) and not prefix):
                    if not package:
                        new_package = name
                    else:
                        new_package = package + '.' + name
                    stack.append((fn, '', new_package, False))
                else:
                    stack.append((fn, prefix + name + '/', package, onlyInPackages))
            elif package or not onlyInPackages:
                # is a file
                bad_name = False
                for pattern in exclude:
                    if (fnmatch.fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()):
                        bad_name = True
                        if showIgnored:
                            print >> sys.stderr, ("File %s ignored by pattern %s" % (fn, pattern))
                        break
                if bad_name:
                    continue
                if len(relativePath):
                    out.setdefault(package, []).append(relativePath+'/'+prefix+name)
                else:
                    out.setdefault(package, []).append(prefix+name)
    return out


##########################################################################################
######################################  EXTENSIONS  ######################################
EXTENSIONS = [# boundary conditions collection
              Extension('fullrmc.Core.boundary_conditions_collection',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"boundary_conditions_collection.pyx")]),
              # reciprocal space
              Extension('fullrmc.Core.reciprocal_space',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"reciprocal_space.pyx")]),
              # pairs distances
              Extension('fullrmc.Core.pairs_distances',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"pairs_distances.pyx")]),
              # pairs histograms
              Extension('fullrmc.Core.pairs_histograms',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"pairs_histograms.pyx")]),
              # atomic_distances
              Extension('fullrmc.Core.atomic_distances',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"atomic_distances.pyx")]),
              # atomic coordination number
              Extension('fullrmc.Core.atomic_coordination',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"atomic_coordination.pyx")]),
              # bonds
              Extension('fullrmc.Core.bonds',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"bonds.pyx")]),
              # angles
              Extension('fullrmc.Core.angles',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"angles.pyx")]),
              # dihedral angles
              Extension('fullrmc.Core.dihedral_angles',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"dihedral_angles.pyx")]),
              # improper angles
              Extension('fullrmc.Core.improper_angles',
                        include_dirs=[np.get_include()],
                        language="c",
                        sources = [os.path.join(EXTENSIONS_PATH,"improper_angles.pyx")]),
              ]
CMDCLASS = {'build_ext' : build_ext}



##########################################################################################
#####################################  PACKAGE DATA  #####################################
# get packages and remove everything that is not fullrmc
docs = os.path.join(PACKAGE_NAME,"docs")

PACKAGES = get_packages(path=PACKAGE_PATH, exclude=(docs, ) )

for package in list(PACKAGES):
    if PACKAGE_NAME not in package:
        PACKAGES.pop(package)

# get package data
PACKAGE_DATA = find_package_data(where=os.path.join(PACKAGE_NAME, "Examples"),
                                 relativePath="Examples",
                                 package='fullrmc',
                                 showIgnored=False)

# metadata
metadata = dict(# package
                name             = PACKAGE_NAME,
                packages         = list(PACKAGES),
                package_dir      = PACKAGES,
                # package data
                package_data     = PACKAGE_DATA,
                # info
                version          = PACKAGE_INFO['__version__'] ,
                author           = "Bachir AOUN",
                author_email     = "fullrmc@gmail.com",
                # Description
                description      = "\n".join(DESCRIPTION),
                long_description = "\n".join(LONG_DESCRIPTION),
                # Extensions
                ext_modules      = EXTENSIONS,
                cmdclass         = CMDCLASS,
                # online
                url              = "http://bachiraoun.github.io/fullrmc",
                download_url     = "https://github.com/bachiraoun/fullrmc",
                # Licence and classifiers
                license          = 'GNU',
                classifiers      = [_f for _f in CLASSIFIERS.split('\n') if _f],
                platforms        = ["Windows", "Linux", "Mac OS-X", "Unix"],
                # Dependent packages (distributions)
                install_requires = ["pysimplelog>=2.0.0",
                                    "pdbparser>=0.1.7",
                                    "pylocker>=2.0.0",
                                    "pyrep>=3.1.0",
                                    "matplotlib>=2.2.3" ], # numpy and cython are also needed, but this is left out for the user to install.
                setup_requires   = [''],
                )



##########################################################################################
#####################################  LAUNCH SETUP  #####################################
setup(**metadata)



##########################################################################################
###################################  BUILD EXTENSIONS  ###################################
#try:
#    from fullrmc.Core.Collection import get_path
#except (Exception, SystemExit) as e:
#    raise "fullrmc not found: %s"%e
#
#try:
#    PYTHON_EXECUTABLE = sys.executable
#    FULLRMC_PATH      = get_path("fullrmc")
#    SETUP_PATH        = os.path.join(FULLRMC_PATH, "Extensions", "setup.py")
#    os.system(PYTHON_EXECUTABLE + " " + SETUP_PATH + " build_ext --inplace")
#except (Exception, SystemExit) as e:
#    raise "Unable to build extension files: %s"%e
#



"""
Must know
=========
- fullrmc supported version is always the newest one. Older version are
  never supported as updating is very simple.
- fullrmc uses pdbparser library to parse input pdb files and to visualize
  system's atomic structure using `VMD <http://www.ks.uiuc.edu/Research/vmd/>`_.
  VMD which stands for Visual Molecular Dynamics is a free software for
  displaying and animating large molecular systems.
- pdbparser in general automatically detects VMD executable when installed.
  fullrmc will work fine without VMD, but make sure it's installed if you
  want to use the convenient visualize method of fullrmc's Engine.
- Unless :class:`.Engine` set_molecules_indexes method is used explicitly,
  molecules are classified by parsing the structure
  `'.pdb' <http://deposit.rcsb.org/adit/docs/pdb_atom_format.html>`_ file
  **Residue name**, **Residue sequence number** and **Segment identifier**
  attributes. Therefore a molecule is the collection of all atoms sharing
  the same three later attributes value.


Installation
============
Here we describe two methods to installing fullrmc.
   #. The first method is by using `pip <https://pip.pypa.io/en/stable/>`_.
      If you don't know what pip is and you are lazy enough not to read through
      the documentation, pip supports installing from PyPI, version control,
      local projects, and directly from distribution files. To install pip
      download `get_pip.py <https://bootstrap.pypa.io/get-pip.py>`_ file
      and run the following from the command line.

      .. code-block:: bash

          python get_pip.py

   #. The second method is by fullrmc's cloning the source code and
      repository from `github <https://github.com/bachiraoun/fullrmc>`_
      and then compiling the extension files in place.


For your general understanding, fullrmc requires the following:
   #. Python (>= 2.7 and 3),
   #. NumPy (lowest version tested is 1.7.1)
   #. cython (lowest version tested is 0.21.1)
   #. matplotlib (lowest version tested is 1.4)
   #. pyrep (lowest version tested is 1.0.2 - 1.0.3 is used starting
      from fullrmc 3.0.0 - 1.0.4 is used starting from fullrmc 3.1.0
      - 3.0.3 is used starting from fullrmc 3.1.0)
   #. pdbparser (lowest version tested is 0.1.2 - 0.1.3 is used starting
      from fullrmc 0.3.0 - 0.1.4 is used starting from fullrmc 1.0.0
      - 0.1.5 is used starting from fullrmc 1.0.1)
   #. pysimplelog (lowest version tested is 0.2.1)

**Installation using pip:**\n
Installing fullrmc with pip is done in two steps.
Firstly install numpy and cython manually and ensure that
they meet with fullrmc's version requirement.

   .. code-block:: bash

       pip install -U "numpy>=1.7.1"
       pip install -U "cython>=0.21.1"

When you already have a working up-do-date installation of numpy and cython,
you can proceed to installing fullrmc. The rest of the dependencies will be
automatically installed and updated while fullrmc is being installed.

   .. code-block:: bash

       pip install fullrmc

**Installation by cloning github repository:**\n
   * Ensure all fullrmc required packages are installed and up-do-date by executing the
     following python script:

    .. code-block:: python

        # check whether all packages are already installed
        from pkg_resources import parse_version as PV
        for name, ver in [('numpy'      ,'1.7.1') ,
                          ('cython'     ,'0.21.1'),
                          ('pyrep'      ,'3.0.0') ,
                          ('pdbparser'  ,'0.1.5') ,
                          ('pysimplelog','0.3.0') ,
                          ('matplotlib' ,'1.4'  )]:
            try:
                lib = __import__(name)
            except:
                print('%s must be installed for fullrmc to run properly'%(name))
            else:
                if PV(lib.__version__) < PV(ver):
                    print('%s installed version %s is below minimum suggested version %s. Updating %s is highly recommended.'%(name, lib.__version__, ver, name))
                else:
                    print('%s is properly installed and minimum version requirement is met.'%(name)


   * Locate python's site-packages by executing the following python script:

     .. code-block:: python

        import os
        os.path.join(os.path.dirname(os.__file__), 'site_packages')

   * Navigate to site_packages folder and clone git repository from command line:

    .. code-block:: bash

       cd .../site_packages
       git clone https://github.com/bachiraoun/fullrmc.git

   * Change directory to .../site_packages/fullrmc/Extensions. Then compile fullrmc
     extensions from command line as the following:

    .. code-block:: bash

       cd .../site_packages/fullrmc/Extensions
       python setup.py build_ext --inplace


Versions
========
Latest stable version source code: https://github.com/bachiraoun/fullrmc \n
Latest pypi installable version: https://pypi.python.org/pypi/fullrmc/ \n


Version 0.1.0:
--------------
This version is not published.

Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~
   #. Selectors:
      :class:`.RandomSelector`
      :class:`.WeightedRandomSelector`
      :class:`.DefinedOrderSelector`
      :class:`.DirectionalOrderSelector`
   #. Constraints:
      :class:`.InterMolecularDistanceConstraint`
      :class:`.IntraMolecularDistanceConstraint`
      :class:`.BondConstraint`
      :class:`.BondsAngleConstraint`
      :class:`.ImproperAngleConstraint`
      :class:`.PairDistributionConstraint`
      :class:`.PairCorrelationConstraint`
   #. Generators:
      :class:`.TranslationGenerator`
      :class:`.TranslationAlongAxisGenerator`
      :class:`.TranslationAlongSymmetryAxisGenerator`
      :class:`.TranslationTowardsCenterGenerator`
      :class:`.TranslationTowardsAxisGenerator`
      :class:`.TranslationTowardsSymmetryAxisGenerator`
      :class:`.TranslationAlongSymmetryAxisPath`
      :class:`.RotationGenerator`
      :class:`.RotationAboutAxisGenerator`
      :class:`.RotationAboutSymmetryAxisGenerator`
      :class:`.RotationAboutSymmetryAxisPath`
      :class:`.DistanceAgitationGenerator`
      :class:`.AngleAgitationGenerator`




`Version 0.2.0 <https://pypi.python.org/pypi/fullrmc/0.2.0>`_:
--------------------------------------------------------------
This is the first officially published version.

New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. Selectors:
      :class:`.SmartRandomSelector`
   #. Constraints:
      :class:`.AtomicCoordinationNumberConstraint`
   #. Generators:
      :class:`.SwapPositionsGenerator`

Known bugs and issues:
~~~~~~~~~~~~~~~~~~~~~~
   #. Boundary conditions handling for all non cubic, tetragonal nor
      orthorhombic systems is prone to distances calculation errors
      caused by pdbparser (<=0.1.2).




`Version 0.3.0 <https://pypi.python.org/pypi/fullrmc/0.3.0>`_:
--------------------------------------------------------------
Starting from this version, pdbparser (>=0.1.3) is used.

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
   #. Periodic boundary conditions handling fixed and separated from pdbparser.
   #. :class:`.InterMolecularDistanceConstraint` squared deviation computation
      changed from square law to absolute value. Therefore small values are not
      under-represented in the total squared deviation calculation.
   #. :class:`.InterMolecularDistanceConstraint` 'flexible' flag added.




`Version 0.3.1 <https://pypi.python.org/pypi/fullrmc/0.3.1>`_:
--------------------------------------------------------------

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
   #. Erroneous wx imports from thfSimulation Example is removed.
   #. :class:`.PairDistributionConstraint` and :class:`.PairCorrelationConstraint`
      check_experimental_data merged and sorted distances bug fixed.
   #. :class:`.PairDistributionConstraint` and :class:`.PairCorrelationConstraint`
      set_limits bug fixed.
   #. :class:`.Engine` add_group method allows list, set, tuples as well as integers.




`Version 0.3.2 <https://pypi.python.org/pypi/fullrmc/0.3.2>`_:
--------------------------------------------------------------

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
   #. :class:`.PairDistributionConstraint` and :class:`.PairCorrelationConstraint`
      binning precision fixed. Histogram compiled C code adjusted accordingly.
   #. :class:`.AtomicCoordinationNumberConstraint` set_coordination_number_definition
      method error message made more clear.
   #. :class:`.AtomicCoordinationNumberConstraint` compiled C++ code boundary conditions
      distances computation fixed for multiple box separation.




`Version 0.3.3 <https://pypi.python.org/pypi/fullrmc/0.3.3>`_:
--------------------------------------------------------------

New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. :class:`.SwapCentersGenerator`




`Version 0.4.0 <https://pypi.python.org/pypi/fullrmc/0.4.0>`_:
--------------------------------------------------------------

New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. :class:`.StructureFactorConstraint` added computing Static Structure Factor
   #. :class:`.ReducedStructureFactorConstraint` added computing Reduced Static
      Structure Factor

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
   #. :class:`.PairDistributionConstraint` and :class:`.PairCorrelationConstraint`
      plot Y axis label units added.
   #. swap example changed to atomicSimulation and S(q) added to the fitting
      along with G(r)
   #. thfSimulation example change to molecularSimulation




`Version 1.0.0 <https://pypi.python.org/pypi/fullrmc/1.0.0>`_:
--------------------------------------------------------------
This is a main version change from 0.x.y to 1.x.y and that's because non-periodic
boundary conditions are now implemented. Starting from this version shape function
correction can be used. For non-periodic boundary conditions, very big box with periodic
boundaries is automatically set and system shape function is estimated to correct for
the fixed :math:`\\rho_{0}` global density approximation.

Some testing were done prior to publishing this version and it was found out that
computing angles for :class:`.BondsAngleConstraint` can in some cases be wrong.
We recommend updating to this version if :class:`.BondsAngleConstraint` is needed
for fitting molecular systems.

Starting from this version, pdbparser (>=0.1.4) is used.

New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. fullrmc.Constraints.Collection package added along with :class:`.ShapeFunction`
   #. Engine numberDensity added. Used to set correct volume and density when
      InfiniteBoundaries are used
   #. fullrmc.Core.Collection.smooth function added. It can be used to smooth
      out any 1D data. This is useful for noise experimental data

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
   #. BondAngles calculation bug fixed
   #. Constraints all squared deviations attributes and methods changed to
      standard error to conserve the generality
   #. Engine chiSquare attributes and methods changed to total standard error
      to conserve the generality
   #. Logger levels added
   #. rebin core method check argument added
   #. :class:`.StructureFactorConstraint` and :class:`.ReducedStructureFactorConstraint`
      experimentalQValues set as property




`Version 1.0.1 <https://pypi.python.org/pypi/fullrmc/1.0.1>`_:
--------------------------------------------------------------
Starting from this version, pdbparser (>=0.1.5) is used. There is no bugs
fixed in this version comparably to version 1.0.0.
This version corrects for cell visualization compatibility between fullrmc
and `VMD <http://www.ks.uiuc.edu/Research/vmd/>`_.
VMD `cell plugin <http://www.ks.uiuc.edu/Research/vmd/plugins/pbctools/>`_
uses a, b, c, unitcell parameters (the lengths of vectors A, B and C the
3D-vectors of the unitcell sides with the convention that A is parallel to
the x-axis) and alpha, beta, gamma (the angles of the unit cell) for
non-orthorhombic unitcells. This version uses cell plugin no more but
draw the boundary conditions vectors directly. Also visualization method
'boxToCenter' flag is added to correct for box visualization shifts by
translating boundary conditions vectors' centre to atom coordinates one.




`Version 1.1.0 <https://pypi.python.org/pypi/fullrmc/1.1.0>`_:
--------------------------------------------------------------
Several levels of abstractions are made allowing multiple core computations
using openmp. Nevertheless more testing is needed and therefore the current
version is published as single core still.
:class:`.AtomicCoordinationNumberConstraint` uses python objects which
makes it very slow, it must be revisited.

New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. :class:`.Constraint` computation cost property added. This is used at engine
      runtime to minimize computations and enhance performance by computing less
      costly constraints first.

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
    #. :class:`.BondsAngleConstraint` computing angle right vector normalization bug fixed.
    #. :class:`.ImproperAngleConstraint` computing improper vector bug fixed.




`Version 1.2.0 <https://pypi.python.org/pypi/fullrmc/1.2.0>`_:
--------------------------------------------------------------
Starting from this version, running fullrmc on multicore and multiprocessor computers
is possible. Also, pysimplelog (>=0.2.1) is used. Isolated system in non-periodic
boundary conditions are treated correctly as in an infinte box using inifinite boundary
conditions instead of a very big box with periodic boundaries.


New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. :class:`.AtomicCoordinationNumberConstraint` completely reimplemented and optimized.
      Coordination number definition is re-formatted allowing any type of core's and
      shell's atoms. This new implementation computes mean coordination number per shell
      definition rather than per atom. Fractional coordination numbers are allowed and
      this implementation benefits from a tremendous performance increase of 1000 times
      compared to the old one.
   #. Engine.run method 'restartPdb' and 'ncore' arguments added.




`Version 1.2.1 <https://pypi.python.org/pypi/fullrmc/1.2.1>`_:
--------------------------------------------------------------
No bugs fixed but some examples corrected.




`Version 2.0.0 <https://pypi.python.org/pypi/fullrmc/2.0.0>`_:
--------------------------------------------------------------
This is a main version change from 1.x.y to 2.x.y. Starting from this version engine
is no more a single file but a `pyrep <http://bachiraoun.github.io/pyrep/>`_
repository. Using a repository engine instead of a single file has many advantages.
The first and more important advantage is the limitation of accessing and saving single
big engine files when the simulated atomic system is big. The second advantage
is the introduction of the concept of engine frames. Frames can be used to build a
fitting history or even better to creating a statistical structural solution rather than
a single structure one which is prone primarily to overfitting and other errors.
Also, repository engine's insure a sort of backwards compatibility and files can be
retrieved when updating fullrmc's versions. Implementation wise, switching from single
file to repository is the ground layer to taking fullrmc to the next level of
super-computing.
Engine repository is not a binary black box file but a directory or a folder that one
can browse. It's advisable not to manually delete or add files or folders to the
repository. Interacting with fullrmc.Engine instance is what one should do to creating
or updating engine's repository.
The transition to fullrmc 2.x.y. should be done seemlessly at the end of the user
except for small modifications that were made and that are listed below.

Modifications:
~~~~~~~~~~~~~~
    #. :class:`.Engine`: Several modification happened to fullrmc.Engine class definition.
        * Instanciating engine arguments have changed. We found that it's a bad practice
          to set the pdb structure, the constraints, the groups and all other attributes
          at the engine instanciation. Therefore, now engine's instanciation takes more
          appropriate arguments such as the engine's repository path, the frames, the
          log file path and whether to delete a found engine at the given path and start
          fresh using freshStart flag.
        * Engine.is_engine method is added to check whether a given path is a
          fullrmc.Engine repository.
        * Engine frames related methods are also added to manipulate and switch between
          frames.
        * Engine.run method takes no more savePath argument, as the path has to be the
          repository. In case one wants to save in a new repository, he should save the
          engine first using Engine.save method and then run the engine using
          Engine.run method.
    #. :class:`.Constraint`: We also found that constraints must not be allowed to change
       engine and therefore Constraint.set_engine method is now deprecated. Also
       Constraints are now instanciated without engine argument. Constraint's engine is
       set automatically when a constraint is added to an engine.

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
    #. abs function changed in all compiled files as it was creating rounding issues on
       some operating systems.
    #. arcos function cliped for floating errors in angles compiled file.
    #. :class:`.InterMolecularDistanceConstraint`: Computing constraint data in function
       single_atomic_distances_dists of compiled atomic_distances fixed.
    #. :class:`.BondConstraint`: bondsMap fixed from causing definition conflicts if set
       multiple times.
    #. :class:`.BondsAngleConstraint`: anglesMap fixed from causing definition conflicts
       if set multiple times.
    #. :class:`.ImproperAngleConstraint`: anglesMap fixed from causing definition
       conflicts if set multiple times.




`Version 3.0.0 <https://pypi.python.org/pypi/fullrmc/3.0.0>`_:
--------------------------------------------------------------
This is a main version change from 2.x.y to 3.x.y where pyrep (>=1.0.3) is used.
Thusfar until versions 2.x.y, system's number of atoms is fixed throughout the whole
simulation. Starting from this version, fullrmc allows dynamically removing atoms
from the system upon fitting. This is a revolutionary functionality enabling the
reproduction of defects in systems.

New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. :class:`.DihedralAngleConstraint` implemented to control dihedral angles.
   #. :class:`.AtomsCollector` implemented allowing the ability of dynamically remove
      and add re-insert atoms to the system.
   #. :class:`.AtomsRemoveGenerator` implemented to generate atoms removing from system
      upon fitting.
   #. :class:`.EmptyGroup` added to host Removes generators.

Modifications:
~~~~~~~~~~~~~~
    #. all cython modules are completely de-pythonized.
    #. all angle constraints respect the same implementation and interface design.




`Version 3.0.1 <https://pypi.python.org/pypi/fullrmc/3.0.1>`_:
--------------------------------------------------------------

New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. Plot method added to all of :class:`.BondConstraint`,
      :class:`.BondsAngleConstraint`, :class:`.DihedralAngleConstraint` and
      :class:`.ImproperAngleConstraint`.

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
    #. :class:`.DihedralAngleConstraint` FRAME_DATA fixed which was preventing
       constraint from saving to pyrep repository.




`Version 3.0.2 <https://pypi.python.org/pypi/fullrmc/3.0.2>`_:
--------------------------------------------------------------

New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. Redundant objects global container mechanism added. This will sometimes reduce
      some engine's attributes size, allowing fast saving and loading of data.

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
    #. :class:`.MoveGeneratorCombinator` bug from latest update fixed.
    #. :class:`.SwapGenerator` is not allowed to be collected or combined with other
       generators. Also swapList will update automatically upon removing and
       releasing atoms.
    #. :class:`.RemoveGenerator` is not allowed to be collected or combined with other
       generators.
    #. MLSelection, refine and explore examples BiasedEngine fixed




`Version 3.1.0 <https://pypi.python.org/pypi/fullrmc/3.1.0>`_:
--------------------------------------------------------------
Starting from this version, it is possible to custom set atoms weight for all
of :class:`.PairDistributionConstraint`, :class:`.PairCorrelationConstraint`,
:class:`.StructureFactorConstraint` and :class:`.ReducedStructureFactorConstraint`.
Also starting from this version, pysimplelog (>=0.3.0) and pyrep (>=1.0.4) are used.

New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. All constraints export data method added.
   #. All constraints plot method added.
   #. :class:`.IntraMolecularDistanceConstraint` added again after being removed in
      early versions.

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
    #. :class:`.Engine` repository ACID flag is changed to False as it's not needed.
    #. :class:`.StructureFactorConstraint` and :class:`.ReducedStructureFactorConstraint`
       atoms removal bug fixed.
    #. Computing standard error for :class:`.IntraMolecularDistanceConstraint` and
       :class:`.InterMolecularDistanceConstraint` is optimized for big systems and
       constraint's definition. Calculation time is thousands of times faster for
       extended definitions.




`Version 3.1.1 <https://pypi.python.org/pypi/fullrmc/3.1.1>`_:
--------------------------------------------------------------
Some bugs are fixed and documentation is revisited.

Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
    #. :class:`.Engine` set_number_density method PeriodicBoundaries bug fixed.
    #. :class:`.StructureFactorConstraint` set_weighting xrays and neutron
       recognized.




`Version 3.1.2 <https://pypi.python.org/pypi/fullrmc/3.1.2>`_:
--------------------------------------------------------------
Raised `issue <https://bachiraoun.github.io/fullrmc/QAForum.html?place=msg%2Ffullrmc%2FUaDZSOdeN28%2FUdnX8yKOBAAJ>`_
bug fixed




`Version 3.2.0 <https://pypi.python.org/pypi/fullrmc/3.2.0>`_:
--------------------------------------------------------------
Raised `issue <https://bachiraoun.github.io/fullrmc/QAForum.html?place=msg%2Ffullrmc%2F4L1SCEbcAi0%2FsOqvPhiSAgAJ>`_
bug fixed




`Version 3.3.0 <https://pypi.python.org/pypi/fullrmc/3.3.0>`_:
--------------------------------------------------------------
Raised `issue <https://bachiraoun.github.io/fullrmc/QAForum.html?place=msg%2Ffullrmc%2FC0Vmt4Qs4bI%2FIu1guEXMAwAJ>`_
bug fixed




`Version 4.0.0 <https://pypi.python.org/pypi/fullrmc/4.0.0>`_:
--------------------------------------------------------------
Starting from this version, fullrmc has become grid computing capable.
multiframe is implemented to generate a statistical solution of multiple
frames instead of a single frame solution. Along with this version,
an instance of fullrmc will be available on a cloud service that can be
used to launch fullrmc and perform multiframe computations. We will update
all user in the Q&A group when and where this service will be available


New Modules and definitions:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   #. MultiframeUtils module added which will be used to launch fullrmc
      on computation grids
   #. ExperimentalConstraints plot and data export methods factored


Fixes and improvements:
~~~~~~~~~~~~~~~~~~~~~~~
    #. :class:`.StructureFactorConstraint` scale factor bug fixed
    #. :class:`.PairCorrelationConstraint` scale factor bug fixed
    #. Engine run method explore bug fixed
    #. Amputation number density bug fixed for InfiniteBoundaryConditions
"""


__version__    = '4.0.1'

__author__     = "Bachir Aoun"

__email__      = "fullrmc@gmail.com"

__forum__      = "https://groups.google.com/forum/#!forum/fullrmc"

__onlinedoc__  = "http://bachiraoun.github.io/fullrmc/"

__repository__ = "https://github.com/bachiraoun/fullrmc"

__pypi__       = "https://pypi.python.org/pypi/fullrmc"



"""
"""
# standard libraries imports
from __future__ import print_function

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, PI, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import is_integer, is_number, generate_random_integer
from ..Core.MoveGenerator import RemoveGenerator


class AtomsRemoveGenerator(RemoveGenerator):
    """
    This generator allows removing single atoms at a time from the system.
    Atoms are randomly picked from atomsList and returned upon calling
    pick_from_list method.

    :Parameters:
        #. group (None, Group): The group instance which is this case must be
           fullrmc EmptyGroup.
        #. maximumCollected (None, Integer): The maximum number allowed of
           atoms to be removed and collected from the engine. This property
           is general to the system and checks engine's collected atoms not
           the number of removed atoms via this generator. If None is given,
           the remover will not check for the number of already removed atoms
           before attempting a remove.
        #. atomsList (None, list,set,tuple,np.ndarray): The list of atoms
           index to chose and remove from.  If None is given, then all atoms
           in system will be used.
    """
    def __init__(self, group=None, maximumCollected=None, atomsList=None):
        super(AtomsRemoveGenerator, self).__init__(group=group,
                                                   maximumCollected=maximumCollected,
                                                   atomsList=atomsList)

    def pick_from_list(self, engine):
        """
        Randomly picks an atom's index from atomsList. This method checks
        and verifies maximumCollected prior to picking and index.

        :Parameters:
            #. engine (Engine): The engine calling the method.

        :Returns:
            #. index (None, np.ndarray): Atom index wrapped in a numpy.ndarray.
               If None is returned than picking is not allowed.
        """
        if self.atomsList is None:
            # use len(engine.pdb) in order to select real index
            index = generate_random_integer(0, len(engine.pdb)-1)
            index = np.array([index], dtype=INT_TYPE)
        else:
            index = np.random.choice(self.atomsList,1)
        # check maximumCollected
        if self.maximumCollected is not None:
            if len(engine._atomsCollector)>=self.maximumCollected:
                index = np.array([], dtype=INT_TYPE)
        # check if already collected
        if len(index):
            if engine._atomsCollector.is_collected(index[0]):
                index = np.array([], dtype=INT_TYPE)
        # return
        return index



"""
This is a C compiled module to compute 
boundary conditions related calculations
"""

from libc.math cimport sqrt
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
#from cython.parallel import prange

# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32


# declare constants
cdef C_FLOAT32 BOX_LENGTH      = 1.0
cdef C_FLOAT32 HALF_BOX_LENGTH = 0.5
cdef C_FLOAT32 FLOAT32_ZERO    = 0.0
cdef C_FLOAT32 FLOAT32_ONE     = 1.0
cdef C_INT32   INT32_ONE       = 1


cdef extern from "math.h":
    C_FLOAT32 floor(C_FLOAT32 x)
    C_FLOAT32 ceil(C_FLOAT32 x)
    C_FLOAT32 sqrt(C_FLOAT32 x)

cdef inline C_FLOAT32 round(C_FLOAT32 num):
    return floor(num + HALF_BOX_LENGTH) if (num > FLOAT32_ZERO) else ceil(num - HALF_BOX_LENGTH)
    



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def get_reciprocal_basis( np.ndarray[C_FLOAT32, ndim=2] basis not None):
    """
    Computes reciprocal box matrix.
    
    :Arguments:
       #. basis (float32 array): The (3,3) box matrix
                                         
    :Returns:
       #. rbasis (float32 array): The (3,3) reciprocal box matrix.
    """
    # declare rbasis
    cdef ndarray[C_FLOAT32,  mode="c", ndim=2] rbasis = np.empty((3,3), dtype=NUMPY_FLOAT32)
    # get reciprocal basis
    rbasis[0,0] = basis[1,1]*basis[2,2] - basis[2,1]*basis[1,2]
    rbasis[0,1] = - ( basis[0,1]*basis[2,2] - basis[2,1]*basis[0,1] )
    rbasis[0,2] = basis[0,1]*basis[1,2] - basis[1,1]*basis[0,2]
    rbasis[1,0] = - ( basis[1,0]*basis[2,2] - basis[2,0]*basis[1,2] )
    rbasis[1,1] = basis[0,0]*basis[2,2] - basis[2,0]*basis[0,2]
    rbasis[1,2] = - ( basis[0,0]*basis[1,2] - basis[1,0]*basis[0,2] )
    rbasis[2,0] = basis[1,0]*basis[2,1] - basis[2,0]*basis[1,1]
    rbasis[2,1] = - ( basis[0,0]*basis[2,1] - basis[2,0]*basis[0,1] )
    rbasis[2,2] = basis[0,0]*basis[1,1] - basis[1,0]*basis[0,1]
    # find volume
    vol = basis[0,0]*rbasis[0,0] + basis[1,0]*rbasis[0,1] + basis[2,0]*rbasis[0,2]
    # normalize with volume
    rbasis[0,0] = rbasis[0,0]/vol
    rbasis[0,1] = rbasis[0,1]/vol
    rbasis[0,2] = rbasis[0,2]/vol
    rbasis[1,0] = rbasis[1,0]/vol
    rbasis[1,1] = rbasis[1,1]/vol
    rbasis[1,2] = rbasis[1,2]/vol
    rbasis[2,0] = rbasis[2,0]/vol
    rbasis[2,1] = rbasis[2,1]/vol
    rbasis[2,2] = rbasis[2,2]/vol
    # return rbasis
    return rbasis, vol
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def transform_coordinates( np.ndarray[C_FLOAT32, ndim=2] transMatrix not None,
                           np.ndarray[C_FLOAT32, ndim=2] coords not None):
    """
    Transforms coordinates array using a transformation matrix.
    
    :Arguments:
       #. transMatrix (float32 array): The (3,3) transformation matrix
       #. coords (float32 array): The (N,3) coordinates array.
                                         
    :Returns:
       #. transCoords (float32 array): The (N,3) transformed coordinates array.
    """
    # declare variables
    cdef C_INT32 i
    # declare transCoords
    cdef ndarray[C_FLOAT32,  mode="c", ndim=2] transCoords = np.empty((coords.shape[0],3), dtype=NUMPY_FLOAT32)
    # loop
    for i from 0 <= i < coords.shape[0]:
        transCoords[i,0] = coords[i,0]*transMatrix[0,0] + coords[i,1]*transMatrix[1,0] + coords[i,2]*transMatrix[2,0]
        transCoords[i,1] = coords[i,0]*transMatrix[0,1] + coords[i,1]*transMatrix[1,1] + coords[i,2]*transMatrix[2,1]
        transCoords[i,2] = coords[i,0]*transMatrix[0,2] + coords[i,1]*transMatrix[1,2] + coords[i,2]*transMatrix[2,2]
    # return transformed
    return transCoords
    


@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def box_coordinates_real_distances( C_INT32                       atomIndex, 
                                    ndarray[C_INT32, ndim=1]      indexes not None,
                                    np.ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                    np.ndarray[C_FLOAT32, ndim=2] basis not None):
                           
    """
    Computes atomic real distances given box coordinates.
    
    :Arguments:
       #. atomIndex (int32): The index of atom to compute the distance from.
       #. indexes (int32 array): The list of atom indexes to compute the distance to
       #. boxCoords (float32 array): The (N,3) box coordinates array.
       #. basis (float32 array): The (3,3) box matrix
       
    :Returns:
       #. distances (float32 array): The (N,) distances array.
    """
    # declare variables
    cdef C_INT32 i, ii
    cdef C_FLOAT32 box_dx, box_dy, box_dz
    cdef C_FLOAT32 real_dx, real_dy, real_dz, distance,
    cdef C_FLOAT32 atomBox_x, atomBox_y, atomBox_z
    # get point coordinates
    atomBox_x = boxCoords[atomIndex,0]
    atomBox_y = boxCoords[atomIndex,1]
    atomBox_z = boxCoords[atomIndex,2]
    # declare transCoords
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] distances = np.empty((indexes.shape[0],), dtype=NUMPY_FLOAT32)
    # loop
    ii = -1
    for i in indexes:
        ii += 1
        # calculate difference
        box_dx = boxCoords[i,0]-atomBox_x
        box_dy = boxCoords[i,1]-atomBox_y
        box_dz = boxCoords[i,2]-atomBox_z
        box_dx -= round(box_dx)
        box_dy -= round(box_dy)
        box_dz -= round(box_dz)
        # get real difference
        real_dx = box_dx*basis[0,0] + box_dy*basis[1,0] + box_dz*basis[2,0]
        real_dy = box_dx*basis[0,1] + box_dy*basis[1,1] + box_dz*basis[2,1]
        real_dz = box_dx*basis[0,2] + box_dy*basis[1,2] + box_dz*basis[2,2]
        # calculate distance         
        distances[ii] = <C_FLOAT32>sqrt(real_dx*real_dx + real_dy*real_dy + real_dz*real_dz)
    # return distances
    return distances
        
        

        


"""
Swaps contains all swap or atoms position exchange MoveGenerator classes.

.. inheritance-diagram:: fullrmc.Generators.Swaps
    :parts: 1

"""
# standard libraries imports
from __future__ import print_function

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import is_number, is_integer
from ..Core.MoveGenerator import  MoveGenerator, SwapGenerator
from ..Core.Collection import _Container

class SwapPositionsGenerator(SwapGenerator):
    """
    Generates positional swapping between atoms of the selected group and
    other atoms randomly selected from swapList.

    :Parameters:
        #. group (None, Group): The group instance.
        #. swapLength (Integer): The swap length that defines the length
           of the group and the length of the every swap sub-list in swapList.
        #. swapList (None, List): The list of atoms.\n
           If None is given, no swapping or exchanging will be performed.\n
           If List is given, it must contain lists of atoms where every
           sub-list must have the same number of atoms as the group.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Swaps import SwapPositionsGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        ##### set swap moves between Lithium and Manganese atoms in Li2MnO3 system #####
        # reset engine groups to atoms to insure atomic grouping of all the system's atoms
        ENGINE.set_groups_as_atoms()
        # get all elements list
        elements = ENGINE.allElements
        # create list of lithium atoms indexes
        liIndexes = [[idx] for idx in xrange(len(elements)) if elements[idx]=='li']
        # create list of manganese atoms indexes
        mnIndexes = [[idx] for idx in xrange(len(elements)) if elements[idx]=='mn']
        # create swap generator to lithium atoms
        swapWithLi = SwapPositionsGenerator(swapList=liIndexes)
        # create swap generator to manganese atoms
        swapWithMn = SwapPositionsGenerator(swapList=mnIndexes)
        # set swap generator to groups
        for g in ENGINE.groups:
            # get group's atom index
            idx = g.indexes[0]
            # set swap to manganese for lithium atoms
            if elements[idx]=='li':
                g.set_move_generator(swapWithMn)
            # set swap to lithium for manganese atoms
            elif elements[idx]=='mn':
                g.set_move_generator(swapWithLi)
            # the rest are oxygen atoms. Default RandomTranslation generator are kept.

    """
    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): The Group instance.
        """
        return True, ""

    def set_swap_length(self, swapLength):
        """
        Set swap length. It will reset swaplist automatically.

        :Parameters:
            #. swapLength (Integer): The swap length that defines the length
               of the group and the length of the every swap sub-list in
               swapList.
        """
        super(SwapPositionsGenerator, self).set_swap_length(swapLength=swapLength)
        self.__swapArray = np.empty( (self.swapLength,3), dtype=FLOAT_TYPE )

    def transform_coordinates(self, coordinates, argument=None):
        """
        Transform coordinates by swapping. This method is called in every move.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the swapping.
            #. argument (object): Any other argument needed to perform
               the move. In General it's not needed.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the move.
        """
        # swap coordinates
        self.__swapArray[:,:] = coordinates[:self.swapLength ,:]
        coordinates[:self.swapLength ,:] = coordinates[self.swapLength :,:]
        coordinates[self.swapLength :,:] = self.__swapArray[:,:]
        # return
        return coordinates


class SwapCentersGenerator(SwapGenerator):
    """
    Computes geometric center of the selected group, and swaps its atoms
    by translation to the atoms geometric center of the other atoms which
    are randomly selected from swapList and vice-versa.

    :Parameters:
        #. group (None, Group): The group instance.
        #. swapList (None, List): The list of atoms.\n
           If None is given, no swapping or exchanging will be performed.\n
           If List is given, it must contain lists of atoms index.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Swaps import SwapCentersGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        ##### set swap moves between first 10 molecular groups of a system #####
        # reset engine groups to molecules
        ENGINE.set_groups_as_molecules()
        # set swap generator to the first 10 groups
        GROUPS = [ENGINE.groups[idx] for idx in range(10)]
        for gidx, group in enumerate(GROUPS):
            swapList = [g.indexes for idx,g in enumerate(GROUPS) if idx!=gidx]
            swapGen  = SwapCentersGenerator(swapList=swapList)
            group.set_move_generator(swapGen)
    """
    def __init__(self, group=None, swapList=None):
        super(SwapCentersGenerator, self).__init__(group=group, swapLength=None, swapList=swapList )

    @property
    def swapLength(self):
        """ Swap length. In this Case it is always None as
        swapLength is not required for this generator."""
        return self.__swapLength

    def set_swap_length(self, swapLength):
        """
        Set swap length. The swap length that defines the length of the
        group and the length of the every swap sub-list in swapList.
        It will automatically be set to None as SwapCentersGenerator
        does not require a fixed length.

        :Parameters:
            #. swapLength (None): The swap length.
        """
        self.__swapLength = None
        self.__swapList   = ()
        # set uncollected atoms swapList
        self._remainingAtomsSwapList = self.__swapList
        # reset collector state
        self._collectorState = None

    def set_group(self, group):
        """
        Set the MoveGenerator group.

        :Parameters:
            #. group (None, Group): group instance.
        """
        MoveGenerator.set_group(self, group)

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): the Group instance.
        """
        return True, ""

    def transform_coordinates(self, coordinates, argument=None):
        """
        Transform coordinates by swapping. This method is called in every move.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the swapping.
            #. argument (object): Any other argument needed to perform the
               move. In General it's not needed.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after
               applying the move.
        """
        # get translation vector
        swapLength    = len(self.groupAtomsIndexes)
        swapsOfCenter = np.mean(coordinates[:swapLength,:], axis=0)
        swapsToCenter = np.mean(coordinates[swapLength :,:], axis=0)
        direction     = swapsToCenter-swapsOfCenter
        # swap by translation
        coordinates[: swapLength,:] += direction
        coordinates[swapLength :,:] -= direction
        # return
        return coordinates
